# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Light curve-related functions."""

import os
import warnings
import copy
from astropy import log
import numpy as np
from astropy.logger import AstropyUserWarning
from stingray.lightcurve import Lightcurve
from stingray.utils import assign_value_if_none
from stingray.gti import create_gti_mask, cross_gtis, contiguous_regions
from .base import (
    _look_for_array_in_array,
    hen_root,
    mkdir_p,
    interpret_bintime,
)
from .io import load_events, save_lcurve, load_lcurve
from .io import HEN_FILE_EXTENSION, high_precision_keyword_read, get_file_type
from .base import deorbit_events


def join_lightcurve_objs(lclist):
    """Join light curves.

    Light curves from different instruments are put in different channels.
    Light curves from the same time interval and instrument raise
    a ValueError.

    Parameters
    ----------
    lclist : list of :class:`Lightcurve` objects
        The list of light curves to join

    Returns
    -------
    lcoutlist : joint light curves, one per instrument

    See Also
    --------
        scrunch_lightcurves : Create a single light curve from input light
                                 curves.

    Examples
    --------
    >>> lcA = Lightcurve(np.arange(4), np.zeros(4))
    >>> lcA.instr='BU'  # Test also case sensitivity
    >>> lcB = Lightcurve(np.arange(4) + 4, [1, 3, 4, 5])
    >>> lcB.instr='bu'
    >>> lcC = join_lightcurve_objs((lcA, lcB))
    >>> np.all(lcC['bu'].time == np.arange(8))
    True
    >>> np.all(lcC['bu'].counts == [0, 0, 0, 0, 1, 3, 4, 5])
    True
    """
    # --------------- Check consistency of data --------------
    lcdts = [lcdata.dt for lcdata in lclist]
    # Find unique elements. If multiple bin times are used, throw an exception
    lcdts = list(set(lcdts))
    assert len(lcdts) == 1, "Light curves must have same dt for joining"

    instrs = [
        lcdata.instr.lower()
        for lcdata in lclist
        if (hasattr(lcdata, "instr") and lcdata.instr is not None)
    ]

    # Find unique elements. A lightcurve will be produced for each instrument
    instrs = list(set(instrs))
    if instrs == []:
        instrs = ["unknown"]

    outlcs = {}
    for instr in instrs:
        outlcs[instr.lower()] = None
    # -------------------------------------------------------

    for lcdata in lclist:
        instr = assign_value_if_none(lcdata.instr, "unknown").lower()
        if outlcs[instr] is None:
            outlcs[instr] = lcdata
        else:
            outlcs[instr] = outlcs[instr].join(lcdata)

    return outlcs


def join_lightcurves(lcfilelist, outfile="out_lc" + HEN_FILE_EXTENSION):
    """Join light curves from different files.

    Light curves from different instruments are put in different channels.

    Parameters
    ----------
    lcfilelist : list of str
        List of input file names
    outfile :
        Output light curve
    See Also
    --------
        scrunch_lightcurves : Create a single light curve from input light
                                 curves.

    """
    lcdatas = []

    for lfc in lcfilelist:
        log.info("Loading file %s..." % lfc)
        lcdata = load_lcurve(lfc)
        log.info("Done.")
        lcdatas.append(lcdata)
        del lcdata

    outlcs = join_lightcurve_objs(lcdatas)

    if outfile is not None:
        instrs = list(outlcs.keys())
        for instr in instrs:
            if len(instrs) == 1:
                tag = ""
            else:
                tag = instr
            log.info("Saving joined light curve to %s" % outfile)

            dname, fname = os.path.split(outfile)
            save_lcurve(outlcs[instr], os.path.join(dname, tag + fname))

    return outlcs


def scrunch_lightcurve_objs(lclist):
    """Create a single light curve from input light curves.

    Light curves are appended when they cover different times, and summed when
    they fall in the same time range. This is done regardless of the channel
    or the instrument.

    Parameters
    ----------
    lcfilelist : list of :class:`stingray.lightcurve.Lightcurve` objects
        The list of light curves to scrunch

    Returns
    -------
    lc : scrunched light curve

    See Also
    --------
        join_lightcurves : Join light curves from different files

    Examples
    --------
    >>> lcA = Lightcurve(np.arange(4), np.ones(4))
    >>> lcA.instr='bu1'
    >>> lcB = Lightcurve(np.arange(4), [1, 3, 4, 5])
    >>> lcB.instr='bu2'
    >>> lcC = scrunch_lightcurve_objs((lcA, lcB))
    >>> np.all(lcC.time == np.arange(4))
    True
    >>> np.all(lcC.counts == [2, 4, 5, 6])
    True
    >>> np.all(lcC.instr == 'bu1,bu2')
    True
    """

    instrs = [lc.instr for lc in lclist]
    gti_lists = [lc.gti for lc in lclist]
    gti = cross_gtis(gti_lists)
    for lc in lclist:
        lc.gti = gti
        if hasattr(lc, "_apply_gtis"):  # pragma: no cover
            # Compatibility with old versions of stingray
            lc.apply_gtis = lc._apply_gtis
        lc.apply_gtis()
    # Determine limits

    lc0 = lclist[0]

    for lc in lclist[1:]:
        lc0 = lc0 + lc

    lc0.instr = ",".join(instrs)

    return lc0


def scrunch_lightcurves(
    lcfilelist, outfile="out_scrlc" + HEN_FILE_EXTENSION, save_joint=False
):
    """Create a single light curve from input light curves.

    Light curves are appended when they cover different times, and summed when
    they fall in the same time range. This is done regardless of the channel
    or the instrument.

    Parameters
    ----------
    lcfilelist : list of str
        The list of light curve files to scrunch

    Returns
    -------
    time : array-like
        The time array
    lc :
        The new light curve
    gti : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        Good Time Intervals

    Other Parameters
    ----------------
    outfile : str
        The output file name
    save_joint : bool
        If True, save the per-channel joint light curves

    See Also
    --------
        join_lightcurves : Join light curves from different files
    """
    if save_joint:
        lcdata = join_lightcurves(lcfilelist)
    else:
        lcdata = join_lightcurves(lcfilelist, outfile=None)

    lc0 = scrunch_lightcurve_objs(list(lcdata.values()))
    log.info("Saving scrunched light curve to %s" % outfile)
    save_lcurve(lc0, outfile)

    return lc0


def filter_lc_gtis(
    lc, safe_interval=None, delete=False, min_length=0, return_borders=False
):
    """Filter a light curve for GTIs.

    Parameters
    ----------
    lc : :class:`Lightcurve` object
        The input light curve

    Returns
    -------
    newlc : :class:`Lightcurve` object
        The output light curve
    borders : [[i0_0, i0_1], [i1_0, i1_1], ...], optional
        The indexes of the light curve corresponding to the borders of the
        GTIs. Returned if return_borders is set to True

    Other Parameters
    ----------------
    safe_interval : float or [float, float]
        Seconds to filter out at the start and end of each GTI. If single
        float, these safe windows are equal, otherwise the two numbers refer
        to the start and end of the GTI respectively
    delete : bool
        If delete is True, the intervals outside of GTIs are filtered out from
        the light curve. Otherwise, they are set to zero.
    min_length : float
        Minimum length of GTI. GTIs below this length will be removed.
    return_borders : bool
        If True, return also the indexes of the light curve corresponding to
        the borders of the GTIs
    """
    mask, newgtis = create_gti_mask(
        lc.time,
        lc.gti,
        return_new_gtis=True,
        safe_interval=safe_interval,
        min_length=min_length,
    )

    nomask = np.logical_not(mask)

    newlc = copy.copy(lc)
    newlc.counts[nomask] = 0
    newlc.gti = newgtis

    if return_borders:
        mask = create_gti_mask(lc.time, newgtis)
        borders = contiguous_regions(mask)
        return newlc, borders
    else:
        return newlc


def lcurve_from_events(
    f,
    safe_interval=0,
    pi_interval=None,
    e_interval=None,
    min_length=0,
    gti_split=False,
    ignore_gtis=False,
    bintime=1.0,
    outdir=None,
    outfile=None,
    noclobber=False,
    deorbit_par=None,
):
    """Bin an event list in a light curve.

    Parameters
    ----------
    f : str
        Input event file name
    bintime : float
        The bin time of the output light curve

    Returns
    -------
    outfiles : list
        List of output light curves

    Other Parameters
    ----------------
    safe_interval : float or [float, float]
        Seconds to filter out at the start and end of each GTI. If single
        float, these safe windows are equal, otherwise the two numbers refer
        to the start and end of the GTI respectively
    pi_interval : [int, int]
        PI channel interval to select. Default None, meaning that all PI
        channels are used
    e_interval : [float, float]
        Energy interval to select (only works if event list is calibrated with
        `calibrate`). Default None
    min_length : float
        GTIs below this length will be filtered out
    gti_split : bool
        If True, create one light curve for each good time interval
    ignore_gtis : bool
        Ignore good time intervals, and get a single light curve that includes
        possible gaps
    outdir : str
        Output directory
    outfile : str
        Output file
    noclobber : bool
        If True, do not overwrite existing files

    """
    log.info("Loading file %s..." % f)
    evdata = load_events(f)
    log.info("Done.")

    deorbit_tag = ""
    if deorbit_par is not None:
        evdata = deorbit_events(evdata, deorbit_par)
        deorbit_tag = "_deorb"

    bintime = np.longdouble(interpret_bintime(bintime))

    tag = ""

    gtis = evdata.gti
    tstart = np.min(gtis)
    tstop = np.max(gtis)
    events = evdata.time
    if hasattr(evdata, "instr") and evdata.instr is not None:
        instr = evdata.instr
    else:
        instr = "unknown"

    if ignore_gtis:
        gtis = np.array([[tstart, tstop]])
        evdata.gtis = gtis

    total_lc = evdata.to_lc(100)
    total_lc.instr = instr

    # Then, apply filters
    if pi_interval is not None and np.all(np.array(pi_interval) > 0):
        pis = evdata.pi
        good = np.logical_and(pis > pi_interval[0], pis <= pi_interval[1])
        events = events[good]
        tag = "_PI%g-%g" % (pi_interval[0], pi_interval[1])
    elif e_interval is not None and np.all(np.array(e_interval) > 0):
        if not hasattr(evdata, "energy") or evdata.energy is None:
            raise ValueError(
                "No energy information is present in the file."
                + " Did you run HENcalibrate?"
            )
        es = evdata.energy
        good = np.logical_and(es > e_interval[0], es <= e_interval[1])
        events = events[good]
        tag = "_E%g-%g" % (e_interval[0], e_interval[1])
    else:
        pass

    if tag != "":
        save_lcurve(
            total_lc,
            hen_root(f) + "_std_lc" + deorbit_tag + HEN_FILE_EXTENSION,
        )

    # Assign default value if None
    outfile = assign_value_if_none(
        outfile, hen_root(f) + tag + deorbit_tag + "_lc"
    )

    # Take out extension from name, if present, then give extension. This
    # avoids multiple extensions
    outfile = outfile.replace(HEN_FILE_EXTENSION, "") + HEN_FILE_EXTENSION
    outdir = assign_value_if_none(outdir, os.path.dirname(os.path.abspath(f)))

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn(
            "File exists, and noclobber option used. Skipping",
            AstropyUserWarning,
        )
        return [outfile]

    lc = Lightcurve.make_lightcurve(
        events,
        bintime,
        tstart=tstart,
        tseg=tstop - tstart,
        mjdref=evdata.mjdref,
        gti=gtis,
    )

    lc.instr = instr
    lc.e_interval = e_interval

    lc = filter_lc_gtis(
        lc, safe_interval=safe_interval, delete=False, min_length=min_length
    )

    if len(lc.gti) == 0:
        warnings.warn(
            "No GTIs above min_length ({0}s) found.".format(min_length)
        )
        return

    lc.header = None
    if hasattr(evdata, "header"):
        lc.header = evdata.header

    if gti_split:
        lcs = lc.split_by_gti()
        outfiles = []

        for ib, l0 in enumerate(lcs):
            local_tag = tag + "_gti{:03d}".format(ib)
            outf = hen_root(outfile) + local_tag + "_lc" + HEN_FILE_EXTENSION
            if noclobber and os.path.exists(outf):
                warnings.warn(
                    "File exists, and noclobber option used. Skipping"
                )
                outfiles.append(outf)
            l0.instr = lc.instr
            l0.header = lc.header

            save_lcurve(l0, outf)
            outfiles.append(outf)
    else:
        log.info("Saving light curve to %s" % outfile)
        save_lcurve(lc, outfile)
        outfiles = [outfile]

    # For consistency in return value
    return outfiles


def lcurve_from_fits(
    fits_file,
    gtistring="GTI",
    timecolumn="TIME",
    ratecolumn=None,
    ratehdu=1,
    fracexp_limit=0.9,
    outfile=None,
    noclobber=False,
    outdir=None,
):
    """
    Load a lightcurve from a fits file and save it in HENDRICS format.

    .. note ::
        FITS light curve handling is still under testing.
        Absolute times might be incorrect depending on the light curve format.

    Parameters
    ----------
    fits_file : str
        File name of the input light curve in FITS format

    Returns
    -------
    outfile : [str]
        Returned as a list with a single element for consistency with
        `lcurve_from_events`

    Other Parameters
    ----------------
    gtistring : str
        Name of the GTI extension in the FITS file
    timecolumn : str
        Name of the column containing times in the FITS file
    ratecolumn : str
        Name of the column containing rates in the FITS file
    ratehdu : str or int
        Name or index of the FITS extension containing the light curve
    fracexp_limit : float
        Minimum exposure fraction allowed
    outfile : str
        Output file name
    noclobber : bool
        If True, do not overwrite existing files
    """
    warnings.warn(
        """WARNING! FITS light curve handling is still under testing.
        Absolute times might be incorrect."""
    )
    # TODO:
    # treat consistently TDB, UTC, TAI, etc. This requires some documentation
    # reading. For now, we assume TDB
    from astropy.io import fits as pf
    from astropy.time import Time
    import numpy as np
    from stingray.gti import create_gti_from_condition

    outfile = assign_value_if_none(outfile, hen_root(fits_file) + "_lc")
    outfile = outfile.replace(HEN_FILE_EXTENSION, "") + HEN_FILE_EXTENSION
    outdir = assign_value_if_none(
        outdir, os.path.dirname(os.path.abspath(fits_file))
    )

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn(
            "File exists, and noclobber option used. Skipping",
            AstropyUserWarning,
        )
        return [outfile]

    lchdulist = pf.open(fits_file)
    lctable = lchdulist[ratehdu].data

    # Units of header keywords
    tunit = lchdulist[ratehdu].header["TIMEUNIT"]

    try:
        mjdref = high_precision_keyword_read(
            lchdulist[ratehdu].header, "MJDREF"
        )
        mjdref = Time(mjdref, scale="tdb", format="mjd")
    except Exception:
        mjdref = None

    try:
        instr = lchdulist[ratehdu].header["INSTRUME"]
    except Exception:
        instr = "EXTERN"

    # ----------------------------------------------------------------
    # Trying to comply with all different formats of fits light curves.
    # It's a madness...
    try:
        tstart = high_precision_keyword_read(
            lchdulist[ratehdu].header, "TSTART"
        )
        tstop = high_precision_keyword_read(lchdulist[ratehdu].header, "TSTOP")
    except Exception:
        raise (Exception("TSTART and TSTOP need to be specified"))

    # For nulccorr lcs this whould work

    timezero = high_precision_keyword_read(
        lchdulist[ratehdu].header, "TIMEZERO"
    )
    # Sometimes timezero is "from tstart", sometimes it's an absolute time.
    # This tries to detect which case is this, and always consider it
    # referred to tstart
    timezero = assign_value_if_none(timezero, 0)

    # for lcurve light curves this should instead work
    if tunit == "d":
        # TODO:
        # Check this. For now, I assume TD (JD - 2440000.5).
        # This is likely wrong
        timezero = Time(2440000.5 + timezero, scale="tdb", format="jd")
        tstart = Time(2440000.5 + tstart, scale="tdb", format="jd")
        tstop = Time(2440000.5 + tstop, scale="tdb", format="jd")
        # if None, use NuSTAR defaulf MJDREF
        mjdref = assign_value_if_none(
            mjdref,
            Time(
                np.longdouble("55197.00076601852"), scale="tdb", format="mjd"
            ),
        )

        timezero = (timezero - mjdref).to("s").value
        tstart = (tstart - mjdref).to("s").value
        tstop = (tstop - mjdref).to("s").value

    if timezero > tstart:
        timezero -= tstart

    time = np.array(lctable.field(timecolumn), dtype=np.longdouble)
    if time[-1] < tstart:
        time += timezero + tstart
    else:
        time += timezero

    try:
        dt = high_precision_keyword_read(lchdulist[ratehdu].header, "TIMEDEL")
        if tunit == "d":
            dt *= 86400
    except Exception:
        warnings.warn(
            "Assuming that TIMEDEL is the median difference between the"
            " light curve times",
            AstropyUserWarning,
        )
        dt = np.median(np.diff(time))

    # ----------------------------------------------------------------
    ratecolumn = assign_value_if_none(
        ratecolumn,
        _look_for_array_in_array(["RATE", "RATE1", "COUNTS"], lctable.names),
    )

    rate = np.array(lctable.field(ratecolumn), dtype=float)

    try:
        rate_e = np.array(lctable.field("ERROR"), dtype=np.longdouble)
    except Exception:
        rate_e = np.zeros_like(rate)

    if "RATE" in ratecolumn:
        rate *= dt
        rate_e *= dt

    try:
        fracexp = np.array(lctable.field("FRACEXP"), dtype=np.longdouble)
    except Exception:
        fracexp = np.ones_like(rate)

    good_intervals = (
        (rate == rate) * (fracexp >= fracexp_limit) * (fracexp <= 1)
    )

    rate[good_intervals] /= fracexp[good_intervals]
    rate_e[good_intervals] /= fracexp[good_intervals]

    rate[np.logical_not(good_intervals)] = 0

    try:
        gtitable = lchdulist[gtistring].data
        gti_list = np.array(
            [
                [a, b]
                for a, b in zip(
                    gtitable.field("START"), gtitable.field("STOP")
                )
            ],
            dtype=np.longdouble,
        )
    except Exception:
        gti_list = create_gti_from_condition(time, good_intervals)

    lchdulist.close()

    lc = Lightcurve(
        time=time,
        counts=rate,
        err=rate_e,
        gti=gti_list,
        mjdref=mjdref.mjd,
        dt=dt,
    )

    lc.instr = instr
    lc.header = lchdulist[ratehdu].header.tostring()

    log.info("Saving light curve to %s" % outfile)
    save_lcurve(lc, outfile)
    return [outfile]


def lcurve_from_txt(
    txt_file, outfile=None, noclobber=False, outdir=None, mjdref=None, gti=None
):
    """
    Load a lightcurve from a text file.

    Parameters
    ----------
    txt_file : str
        File name of the input light curve in text format. Assumes two columns:
        time, counts. Times are seconds from MJDREF 55197.00076601852 (NuSTAR)
        if not otherwise specified.

    Returns
    -------
    outfile : [str]
        Returned as a list with a single element for consistency with
        `lcurve_from_events`

    Other Parameters
    ----------------
    outfile : str
        Output file name
    noclobber : bool
        If True, do not overwrite existing files
    mjdref : float, default 55197.00076601852
        the MJD time reference
    gti : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        Good Time Intervals
    """
    import numpy as np

    if mjdref is None:
        mjdref = np.longdouble("55197.00076601852")

    outfile = assign_value_if_none(outfile, hen_root(txt_file) + "_lc")
    outfile = outfile.replace(HEN_FILE_EXTENSION, "") + HEN_FILE_EXTENSION

    outdir = assign_value_if_none(
        outdir, os.path.dirname(os.path.abspath(txt_file))
    )

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn(
            "File exists, and noclobber option used. Skipping",
            AstropyUserWarning,
        )
        return [outfile]

    time, counts = np.genfromtxt(txt_file, delimiter=" ", unpack=True)
    time = np.array(time, dtype=np.longdouble)
    counts = np.array(counts, dtype=float)

    lc = Lightcurve(time=time, counts=counts, gti=gti, mjdref=mjdref)

    lc.instr = "EXTERN"

    log.info("Saving light curve to %s" % outfile)
    save_lcurve(lc, outfile)
    return [outfile]


def _baseline_lightcurves(lcurves, outroot, p, lam):
    outroot_save = outroot
    for i, f in enumerate(lcurves):
        if outroot is None:
            outroot = hen_root(f) + "_lc_baseline"
        else:
            outroot = outroot_save + "_{}".format(i)
        ftype, lc = get_file_type(f)
        baseline = lc.baseline(p, lam)
        lc.base = baseline
        save_lcurve(lc, outroot + HEN_FILE_EXTENSION)


def _wrap_lc(args):
    f, kwargs = args
    try:
        return lcurve_from_events(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        raise


def _wrap_txt(args):
    f, kwargs = args
    try:
        return lcurve_from_txt(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        return []


def _wrap_fits(args):
    f, kwargs = args
    try:
        return lcurve_from_fits(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        return []


def _execute_lcurve(args):
    from multiprocessing import Pool

    bintime = args.bintime

    safe_interval = args.safe_interval
    e_interval, pi_interval = args.energy_interval, args.pi_interval
    if args.pi_interval is not None:
        pi_interval = np.array(args.pi_interval)
    if e_interval is not None:
        args.e_interval = np.array(args.energy_interval)

    # ------ Use functools.partial to wrap lcurve* with relevant keywords---
    if args.fits_input:
        wrap_fun = _wrap_fits
        argdict = {"noclobber": args.noclobber}
    elif args.txt_input:
        wrap_fun = _wrap_txt
        argdict = {"noclobber": args.noclobber}
    else:
        wrap_fun = _wrap_lc
        argdict = {
            "noclobber": args.noclobber,
            "safe_interval": safe_interval,
            "pi_interval": pi_interval,
            "e_interval": e_interval,
            "min_length": args.minlen,
            "gti_split": args.gti_split,
            "ignore_gtis": args.ignore_gtis,
            "bintime": bintime,
            "outdir": args.outdir,
            "deorbit_par": args.deorbit_par,
        }

    arglist = [[f, argdict.copy()] for f in args.files]
    na = len(arglist)
    outfile = args.outfile
    if outfile is not None:
        outname = os.path.splitext(outfile)[0]
        for i in range(na):
            if na > 1:
                outname = outfile + "_{0}".format(i)
            arglist[i][1]["outfile"] = outname

    # -------------------------------------------------------------------------
    outfiles = []

    if os.name == "nt" or args.nproc == 1:
        for a in arglist:
            outfiles.append(wrap_fun(a))
    else:
        pool = Pool(processes=args.nproc)
        for i in pool.imap_unordered(wrap_fun, arglist):
            outfiles.append(i)
        pool.close()

    log.debug(f"{outfiles}")

    if args.scrunch:
        scrunch_lightcurves(outfiles)

    if args.join:
        join_lightcurves(outfiles)


def main(args=None):
    """Main function called by the `HENlcurve` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Create lightcurves starting from event files. It is "
        "possible to specify energy or channel filtering options"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs="+")

    parser.add_argument(
        "-b",
        "--bintime",
        type=float,
        default=1,
        help="Bin time; if negative, negative power of 2",
    )
    parser.add_argument(
        "--safe-interval",
        nargs=2,
        type=float,
        default=[0, 0],
        help="Interval at start and stop of GTIs used" + " for filtering",
    )
    parser = _add_default_args(parser, ["energies", "pi"])
    parser.add_argument(
        "-s",
        "--scrunch",
        help="Create scrunched light curve (single channel)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-j",
        "--join",
        help="Create joint light curve (multiple channels)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-g",
        "--gti-split",
        help="Split light curve by GTI",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--minlen",
        help="Minimum length of acceptable GTIs (default:4)",
        default=4,
        type=float,
    )
    parser.add_argument(
        "--ignore-gtis", help="Ignore GTIs", default=False, action="store_true"
    )
    parser.add_argument(
        "-d", "--outdir", type=str, default=None, help="Output directory"
    )
    parser.add_argument(
        "--noclobber",
        help="Do not overwrite existing files",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--fits-input",
        help="Input files are light curves in FITS format",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--txt-input",
        help="Input files are light curves in txt format",
        default=False,
        action="store_true",
    )
    parser = _add_default_args(
        parser, ["deorbit", "output", "loglevel", "debug", "nproc"]
    )

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)
    if args.debug:
        args.loglevel = "DEBUG"
    log.setLevel(args.loglevel)

    with log.log_to_file("HENlcurve.log"):
        _execute_lcurve(args)


def scrunch_main(args=None):
    """Main function called by the `HENscrunchlc` command line script."""
    import argparse

    description = "Sum lightcurves from different instruments or energy ranges"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        default="out_scrlc" + HEN_FILE_EXTENSION,
        help="Output file",
    )
    parser.add_argument(
        "--loglevel",
        help=(
            "use given logging level (one between INFO, "
            "WARNING, ERROR, CRITICAL, DEBUG; "
            "default:WARNING)"
        ),
        default="WARNING",
        type=str,
    )
    parser.add_argument(
        "--debug",
        help="use DEBUG logging level",
        default=False,
        action="store_true",
    )

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENscrunchlc.log"):
        scrunch_lightcurves(files, args.out)


def baseline_main(args=None):
    """Main function called by the `HENbaselinesub` command line script."""
    import argparse

    description = (
        "Subtract a baseline from the lightcurve using the Asymmetric Least "
        "Squares algorithm. The two parameters p and lambda control the "
        "asymmetry and smoothness of the baseline. See below for details."
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-o", "--out", type=str, default=None, help="Output file"
    )
    parser.add_argument(
        "--loglevel",
        help=(
            "use given logging level (one between INFO, "
            "WARNING, ERROR, CRITICAL, DEBUG; "
            "default:WARNING)"
        ),
        default="WARNING",
        type=str,
    )
    parser.add_argument(
        "--debug",
        help="use DEBUG logging level",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-p",
        "--asymmetry",
        type=float,
        help='"asymmetry" parameter. Smaller values make the '
        'baseline more "horizontal". Typically '
        "0.001 < p < 0.1, but not necessarily.",
        default=0.01,
    )
    parser.add_argument(
        "-l",
        "--lam",
        type=float,
        help='lambda, or "smoothness", parameter. Larger'
        " values make the baseline stiffer. Typically "
        "1e2 < lam < 1e9",
        default=1e5,
    )

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENbaseline.log"):
        _baseline_lightcurves(files, args.out, args.asymmetry, args.lam)



import copy
import os
from functools import wraps
import numpy as np
from astropy.table import Table
from astropy import log
import stingray.utils
from stingray.events import EventList

try:
    from numba import jit, njit, prange, vectorize
    from numba import float32, float64, int32, int64
    from numba import types
    from numba.extending import overload_method

    @overload_method(types.Array, "take")  # pragma: no cover
    def array_take(arr, indices):
        """Adapt np.take to arrays"""
        if isinstance(indices, types.Array):

            def take_impl(arr, indices):
                n = indices.shape[0]
                res = np.empty(n, arr.dtype)
                for i in range(n):
                    res[i] = arr[indices[i]]
                return res

            return take_impl

    HAS_NUMBA = True
except ImportError:
    HAS_NUMBA = False

    def njit(**kwargs):
        """Dummy decorator in case jit cannot be imported."""

        def true_decorator(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                r = func(*args, **kwargs)
                return r

            return wrapped

        return true_decorator

    jit = njit

    def prange(*args):
        """Dummy decorator in case jit cannot be imported."""
        return range(*args)

    class vectorize(object):
        def __init__(self, *args, **kwargs):
            pass

        def __call__(self, func):
            wrapped_f = np.vectorize(func)

            return wrapped_f

    float32 = float64 = int32 = int64 = lambda x, y: None

    def array_take(arr, indices):
        """Adapt np.take to arrays"""
        return np.take(arr, indices)


class _MonkeyPatchedEventList(EventList):
    def apply_mask(self, mask, inplace=False):  # pragma: no cover
        """For compatibility with old stingray version.
        Examples
        --------
        >>> evt = _MonkeyPatchedEventList(time=[0, 1, 2])
        >>> newev0 = evt.apply_mask([True, True, False], inplace=False);
        >>> newev1 = evt.apply_mask([True, True, False], inplace=True);
        >>> np.allclose(newev0.time, [0, 1])
        True
        >>> np.allclose(newev1.time, [0, 1])
        True
        >>> evt is newev1
        True
        """
        if inplace:
            new_ev = self
        else:
            new_ev = copy.deepcopy(self)
        for attr in "time", "energy", "pi", "cal_pi", "detector_id":
            if hasattr(new_ev, attr) and getattr(new_ev, attr) is not None:
                setattr(new_ev, attr, getattr(new_ev, attr)[mask])
        return new_ev



from astropy.io.fits import Header
from stingray.lightcurve import Lightcurve
from stingray.events import EventList
import numpy as np
from hendrics.io import save_events, HEN_FILE_EXTENSION, load_folding
from hendrics.efsearch import main_zsearch
from hendrics.phaseogram import main_phaseogram, run_interactive_phaseogram
from hendrics.phaseogram import InteractivePhaseogram, BinaryPhaseogram
from hendrics.base import hen_root
from hendrics.fold import HAS_PINT
from hendrics.plot import plot_folding
import os
import pytest
import subprocess as sp
from astropy.tests.helper import remote_data


class TestPhaseogram:
    def setup_class(cls):
        cls.pulse_frequency = 1 / 0.101
        cls.tstart = 0
        cls.tend = 25.25
        cls.tseg = cls.tend - cls.tstart
        cls.dt = 0.00606
        cls.times = np.arange(cls.tstart, cls.tend, cls.dt) + cls.dt / 2
        cls.counts = 100 + 20 * np.cos(
            2 * np.pi * cls.times * cls.pulse_frequency
        )
        lc = Lightcurve(
            cls.times, cls.counts, gti=[[cls.tstart, cls.tend]], dt=cls.dt
        )
        events = EventList()
        events.simulate_times(lc)
        events.mjdref = 57000.0
        cls.event_times = events.time
        cls.dum = "events" + HEN_FILE_EXTENSION
        cls.dum_nohead = "events_nohead" + HEN_FILE_EXTENSION
        cls.dum_info = "events_info" + HEN_FILE_EXTENSION
        save_events(events, cls.dum_nohead)

        header = Header()
        events.header = header.tostring()

        save_events(events, cls.dum)
        header = Header()
        header["OBJECT"] = "BUBU"
        header["RADECSYS"] = "FK5"
        header["RA_OBJ"] = 0.4
        header["DEC_OBJ"] = 30.4
        events.header = header.tostring()
        save_events(events, cls.dum_info)

        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        fits_file = os.path.join(cls.datadir, "monol_testA.evt")
        command = "HENreadevents {0}".format(fits_file)
        sp.check_call(command.split())

        cls.real_event_file = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )

    def test_zsearch(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fit-candidates",
                "--fit-frequency",
                str(self.pulse_frequency),
            ]
        )
        outfile = "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        # Defaults to 2 harmonics
        assert efperiod.N == 2

    @pytest.mark.parametrize("label", ("", "_info", "_nohead"))
    def test_phaseogram_input_periodogram(self, label):
        evfile = getattr(self, "dum" + label)
        main_phaseogram(
            [
                evfile,
                "--periodogram",
                "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION,
                "--test",
            ]
        )

    @pytest.mark.parametrize(
        "norm", ["to1", "mediansub", "mediannorm", "meansub", "meannorm"]
    )
    def test_phaseogram_input_norm(self, norm):
        evfile = self.dum
        main_phaseogram(
            [
                evfile,
                "--periodogram",
                "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION,
                "--test",
                "--norm",
                norm,
            ]
        )

    def test_phaseogram_all_defaults(self):
        times = [0, 3.0, 4.0]
        with pytest.warns(UserWarning) as record:
            _ = InteractivePhaseogram(times, 1.0, test=True)
        assert np.any(["MJDREF not set." in r.message.args[0] for r in record])

    def test_binary_all_no_orb(self):
        times = [0, 3.0, 4.0]
        with pytest.raises(RuntimeError) as excinfo:
            # Missing t0
            _ = BinaryPhaseogram(times, 1.0, orbital_period=3.0, asini=3.0)
        assert "Please specify all binary parameters" in str(excinfo.value)

    def test_phaseogram_input_norm_invalid(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            main_phaseogram(
                [
                    evfile,
                    "--periodogram",
                    "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION,
                    "--test",
                    "--norm",
                    "arsdfajl",
                ]
            )
            assert np.any(
                [
                    "Profile normalization arsdfajl" in r.message.args[0]
                    for r in record
                ]
            )

    def test_phaseogram_input_f(self):
        evfile = self.dum
        main_phaseogram([evfile, "-f", "9.9", "--test", "--pepoch", "57000"])

    def test_phaseogram_input_real_data(self):
        evfile = self.real_event_file
        main_phaseogram([evfile, "-f", "9.9", "--test", "--pepoch", "57000"])

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_phaseogram_input_f_change(self):
        evfile = self.dum
        ip = run_interactive_phaseogram(evfile, 9.9, test=True, nbin=16, nt=8)
        ip.update(1)
        ip.recalculate(1)
        with pytest.warns(UserWarning) as record:
            ip.toa(1)
        assert np.any(
            [
                "TOA calculation is not robust" in r.message.args[0]
                for r in record
            ]
        )
        ip.reset(1)
        ip.fdot = 2
        f, fdot, fddot = ip.get_values()
        assert fdot == 2
        assert f == 9.9
        par = hen_root(evfile) + ".par"
        assert os.path.exists(par)

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_phaseogram_deorbit(self):
        evfile = self.dum

        par = hen_root(evfile) + ".par"
        with open(par, "a") as fobj:
            print("F0  9.9", file=fobj)
            print("PEPOCH 57000", file=fobj)
            print("BINARY BT", file=fobj)
            print("PB  1e20", file=fobj)
            print("A1  0", file=fobj)
            print("T0  56000", file=fobj)
            print("EPHEM  DE200", file=fobj)
            print("RAJ  00:55:01", file=fobj)
            print("DECJ 12:00:40.2", file=fobj)

        ip = run_interactive_phaseogram(
            evfile, 9.9, test=True, nbin=16, nt=8, deorbit_par=par
        )
        ip.update(1)
        ip.recalculate(1)
        with pytest.warns(UserWarning) as record:
            ip.toa(1)
        assert np.any(
            [
                "TOA calculation is not robust" in r.message.args[0]
                for r in record
            ]
        )
        ip.reset(1)
        ip.fdot = 2
        f, fdot, fddot = ip.get_values()
        assert fdot == 2
        assert f == 9.9

    def test_phaseogram_raises(self):
        evfile = self.dum
        with pytest.raises(ValueError):
            main_phaseogram([evfile, "--test"])

    def test_phaseogram_input_periodogram_binary(self):
        evfile = self.dum
        main_phaseogram(
            [
                evfile,
                "--binary",
                "--periodogram",
                "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION,
                "--test",
                "--pepoch",
                "57000",
            ]
        )

    def test_phaseogram_input_f_binary(self):
        evfile = self.dum
        main_phaseogram(
            [
                evfile,
                "--binary",
                "-f",
                "9.9",
                "--test",
                "--binary-parameters",
                "10000",
                "0",
                "0",
            ]
        )

    def test_phaseogram_input_f_change_binary(self):
        evfile = self.dum
        ip = run_interactive_phaseogram(evfile, 9.9, test=True, binary=True)
        ip.update(1)
        ip.recalculate(1)
        ip.reset(1)
        ip.zoom_in(1)
        ip.zoom_out(1)
        with pytest.warns(UserWarning) as record:
            ip.toa(1)
        assert np.any(
            [
                "This function was not implemented" in r.message.args[0]
                for r in record
            ]
        )
        ip.orbital_period = 2
        orbital_period, fdot, fddot = ip.get_values()
        assert orbital_period == 2

    def test_phaseogram_raises_binary(self):
        evfile = self.dum
        with pytest.raises(ValueError):
            main_phaseogram([evfile, "--binary", "--test"])

    @classmethod
    def teardown_class(cls):
        os.unlink(cls.dum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test a full run of the codes from the command line."""

import shutil
import os
import glob
import subprocess as sp

import numpy as np
from astropy import log
from astropy.tests.helper import catch_warnings
from astropy.logger import AstropyUserWarning
from astropy.tests.helper import remote_data
import pytest
from stingray.lightcurve import Lightcurve
import hendrics as hen
from hendrics.tests import _dummy_par
from hendrics.fold import HAS_PINT
from hendrics import (
    fake,
    fspec,
    base,
    binary,
    calibrate,
    colors,
    create_gti,
    exposure,
    exvar,
    io,
    lcurve,
    modeling,
    plot,
    read_events,
    rebin,
    save_as_xspec,
    timelags,
    varenergy,
    sum_fspec,
)

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

log.setLevel("DEBUG")
# log.basicConfig(filename='HEN.log', level=log.DEBUG, filemode='w')


def find_file_pattern_in_dir(pattern, directory):
    return glob.glob(os.path.join(directory, pattern))


class TestFullRun(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.par = _dummy_par("bubububu.par")

        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileB = os.path.join(
            cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileAcal = os.path.join(
            cls.datadir,
            "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.ev_fileBcal = os.path.join(
            cls.datadir,
            "monol_testB_nustar_fpmb_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.par = _dummy_par("bubububu.par")
        command = "{0} {1} --discard-calibration".format(
            os.path.join(cls.datadir, "monol_testA.evt"),
            os.path.join(cls.datadir, "monol_testB.evt"),
        )
        hen.read_events.main(command.split())
        command = "{} {} -r {}".format(
            os.path.join(
                cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(
                cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(cls.datadir, "test.rmf"),
        )
        hen.calibrate.main(command.split())
        cls.lcA = os.path.join(
            os.path.join(
                cls.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        cls.lcB = os.path.join(
            os.path.join(
                cls.datadir, "monol_testB_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        command = (
            "{} -e 3 50 --safe-interval 100 300  --nproc 2 -b 0.5 " "-o {}"
        ).format(cls.ev_fileAcal, cls.lcA)
        hen.lcurve.main(command.split())
        command = (
            "{} -e 3 50 --safe-interval 100 300  --nproc 2 -b 0.5 " "-o {}"
        ).format(cls.ev_fileBcal, cls.lcB)
        hen.lcurve.main(command.split())

        cls.pdsA = os.path.join(
            cls.datadir, "monol_testA_E3-50_pds" + HEN_FILE_EXTENSION
        )
        cls.pdsB = os.path.join(
            cls.datadir, "monol_testB_E3-50_pds" + HEN_FILE_EXTENSION
        )
        cls.cpds = os.path.join(
            cls.datadir, "monol_test_E3-50_cpds" + HEN_FILE_EXTENSION
        )

        command = "{} {} -f 128 -k PDS --save-all --norm leahy".format(
            cls.lcA, cls.lcB
        )
        hen.fspec.main(command.split())

        command = "{} {} -f 128 -k CPDS --save-all --norm leahy".format(
            cls.lcA, cls.lcB
        )
        hen.fspec.main(command.split())
        assert os.path.exists(cls.cpds)
        assert os.path.exists(cls.pdsA)
        assert os.path.exists(cls.pdsB)

    def test_scripts_are_installed(self):
        """Test only once that command line scripts are installed correctly."""
        fits_file = os.path.join(self.datadir, "monol_testA.evt")
        command = "HENreadfile {0}".format(fits_file)
        sp.check_call(command.split())

    def test_get_file_type(self):
        """Test getting file type."""
        file_list = {
            "events": "monol_testA_nustar_fpma_ev",
            "lc": "monol_testA_E3-50_lc",
            "pds": "monol_testA_E3-50_pds",
            "cpds": "monol_test_E3-50_cpds",
        }
        for realtype in file_list.keys():
            fname = os.path.join(
                self.datadir, file_list[realtype] + HEN_FILE_EXTENSION
            )
            ftype, _ = hen.io.get_file_type(fname)
            assert ftype == realtype, "File types do not match"

    def test_save_varen_rms(self):
        fname = self.ev_fileAcal
        hen.varenergy.main(
            [
                fname,
                "-f",
                "0",
                "100",
                "--energy-values",
                "0.3",
                "12",
                "5",
                "lin",
                "--rms",
                "-b",
                "0.5",
                "--segment-size",
                "128",
            ]
        )
        out = hen.base.hen_root(fname) + "_rms" + ".qdp"
        os.path.exists(out)

    def test_save_varen_rms(self):
        fname = self.ev_fileAcal
        hen.varenergy.main(
            [
                fname,
                "-f",
                "0",
                "100",
                "--energy-values",
                "0.3",
                "12",
                "5",
                "lin",
                "--covariance",
                "-b",
                "0.5",
                "--segment-size",
                "128",
            ]
        )
        out = hen.base.hen_root(fname) + "_cov" + ".qdp"
        os.path.exists(out)

    def test_save_varen_lag(self):
        fname = self.ev_fileAcal
        hen.varenergy.main(
            [
                fname,
                "-f",
                "0",
                "100",
                "--energy-values",
                "0.3",
                "12",
                "5",
                "lin",
                "--lag",
                "-b",
                "0.5",
                "--segment-size",
                "128",
            ]
        )
        out = hen.base.hen_root(fname) + "_lag" + ".qdp"
        os.path.exists(out)

    def test_colors_fail_uncalibrated(self):
        """Test light curve using PI filtering."""
        command = ("{0} -b 100 -e {1} {2} {2} {3}").format(
            self.ev_fileA, 3, 5, 10
        )
        with catch_warnings():
            with pytest.raises(ValueError) as excinfo:
                hen.colors.main(command.split())

        assert "No energy information is present " in str(excinfo.value)

    def test_colors(self):
        """Test light curve using PI filtering."""
        # calculate colors
        command = ("{0} -b 100 -e {1} {2} {2} {3}").format(
            self.ev_fileAcal, 3, 5, 10
        )
        hen.colors.main(command.split())

        new_filename = os.path.join(
            self.datadir,
            "monol_testA_nustar_fpma_E_10-5_over_5-3" + HEN_FILE_EXTENSION,
        )

        assert os.path.exists(new_filename)
        out_lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.ev_fileA).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def test_readfile_fits(self):
        """Test reading and dumping a FITS file."""
        fitsname = os.path.join(self.datadir, "monol_testA.evt")
        command = "{0}".format(fitsname)

        hen.io.main(command.split())

    def test_plot_color(self):
        """Test plotting with linear axes."""
        lname = (
            os.path.join(
                self.datadir, "monol_testA_nustar_fpma_E_10-5_over_5-3"
            )
            + HEN_FILE_EXTENSION
        )
        cname = (
            os.path.join(
                self.datadir, "monol_testA_nustar_fpma_E_10-5_over_5-3"
            )
            + HEN_FILE_EXTENSION
        )
        hen.plot.main(
            [
                cname,
                lname,
                "--noplot",
                "--xlog",
                "--ylog",
                "--CCD",
                "-o",
                "dummy.qdp",
            ]
        )

    def test_plot_hid(self):
        """Test plotting with linear axes."""
        # also produce a light curve with the same binning
        command = ("{0} -b 100 --energy-interval {1} {2}").format(
            os.path.join(
                self.datadir,
                "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
            ),
            3,
            10,
        )

        hen.lcurve.main(command.split())
        lname = (
            os.path.join(self.datadir, "monol_testA_nustar_fpma_E3-10_lc")
            + HEN_FILE_EXTENSION
        )
        os.path.exists(lname)
        cname = (
            os.path.join(
                self.datadir, "monol_testA_nustar_fpma_E_10-5_over_5-3"
            )
            + HEN_FILE_EXTENSION
        )
        hen.plot.main(
            [
                cname,
                lname,
                "--noplot",
                "--xlog",
                "--ylog",
                "--HID",
                "-o",
                "dummy.qdp",
            ]
        )

    def test_all_files_get_read(self):
        # Test that HENreadfile works with all file types
        fnames = find_file_pattern_in_dir(
            "*" + HEN_FILE_EXTENSION, self.datadir
        )
        io.main(fnames)

    @classmethod
    def teardown_class(self):
        """Test a full run of the scripts (command lines)."""

        patterns = [
            "*monol_test*" + HEN_FILE_EXTENSION,
            "*lcurve*" + HEN_FILE_EXTENSION,
            "*lcurve*.txt",
            "*.log",
            "*monol_test*.dat",
            "*monol_test*.png",
            "*monol_test*.txt",
            "*monol_test_fake*.evt",
            "*bubu*",
            "*.p",
            "*.qdp",
            "*.inf",
        ]

        file_list = []
        for pattern in patterns:
            file_list.extend(find_file_pattern_in_dir(pattern, self.datadir))

        for f in file_list:
            if os.path.exists(f):
                print("Removing " + f)
                os.remove(f)

        patterns = ["*_pds*/", "*_cpds*/", "*_sum/"]

        dir_list = []
        for pattern in patterns:
            dir_list.extend(find_file_pattern_in_dir(pattern, self.datadir))
        for f in dir_list:
            if os.path.exists(f):
                shutil.rmtree(f)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import glob
import pytest
import numpy as np
from astropy.tests.helper import catch_warnings

from stingray.events import EventList
from hendrics.read_events import treat_event_file
from hendrics.io import HEN_FILE_EXTENSION, load_data, save_events, load_events
from hendrics.io import ref_mjd
from hendrics.io import main as main_readfile
from hendrics.fake import main
import hendrics as hen


class TestMergeEvents:
    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")

        ev0 = EventList(
            time=np.sort(np.random.uniform(0, 100, 10)),
            gti=np.array([[0.0, 100]]),
            mjdref=1,
        )
        ev1 = EventList(
            time=np.sort(np.random.uniform(200, 300, 10)),
            gti=np.array([[200.0, 300]]),
            mjdref=2,
        )
        ev2 = EventList(
            time=np.sort(np.random.uniform(400, 500, 10)),
            gti=np.array([[400.0, 500]]),
            mjdref=1,
        )
        ev3 = EventList(
            time=np.sort(np.random.uniform(600, 700, 10)),
            gti=np.array([[600.0, 700]]),
            mjdref=1,
        )
        ev4 = EventList(
            time=np.sort(np.random.uniform(0, 100, 10)),
            gti=np.array([[600.0, 700]]),
            mjdref=1,
        )

        for ev in [ev0, ev1, ev2, ev3]:
            ev.pi = np.random.randint(0, 10, ev.time.size)
            ev.energy = np.random.uniform(3, 79, ev.time.size)

        ev0.instr = ev1.instr = ev2.instr = ev4.instr = "BA"
        ev3.instr = "BU"
        f0 = os.path.join(cls.datadir, "ev0_ev" + HEN_FILE_EXTENSION)
        f1 = os.path.join(cls.datadir, "ev1_ev" + HEN_FILE_EXTENSION)
        f2 = os.path.join(cls.datadir, "ev2_ev" + HEN_FILE_EXTENSION)
        f3 = os.path.join(cls.datadir, "ev3_ev" + HEN_FILE_EXTENSION)
        f4 = os.path.join(cls.datadir, "ev4_ev" + HEN_FILE_EXTENSION)

        save_events(ev0, f0)
        save_events(ev1, f1)
        save_events(ev2, f2)
        save_events(ev3, f3)
        save_events(ev4, f4)

        cls.f0, cls.f1, cls.f2, cls.f3, cls.f4 = f0, f1, f2, f3, f4

    def test_merge_events(self):
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join(
                [
                    self.f0,
                    self.f1,
                    "-o",
                    os.path.join(
                        self.datadir, "monol_merg_ev" + HEN_FILE_EXTENSION
                    ),
                ]
            )
        assert np.any(["changing MJDREF" in r.message.args[0] for r in record])

        out = os.path.join(self.datadir, "monol_merg_ev" + HEN_FILE_EXTENSION)
        assert os.path.exists(out)
        os.unlink(out)

    def test_merge_events_different_instr(self):
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join(
                [
                    self.f0,
                    self.f3,
                    "-o",
                    os.path.join(
                        self.datadir, "monol_merg13_ev" + HEN_FILE_EXTENSION
                    ),
                ]
            )

        out = os.path.join(
            self.datadir, "monol_merg13_ev" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(out)
        os.unlink(out)

    def test_merge_events_no_out_fname(self):
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join([self.f0, self.f1])
        assert np.any(["changing MJDREF" in r.message.args[0] for r in record])
        out = os.path.join(self.datadir, "ev_ev" + HEN_FILE_EXTENSION)
        assert os.path.exists(out)
        os.unlink(out)

    def test_merge_many_events_warnings(self):

        out = os.path.join(
            self.datadir, "monol_merg_many_ev" + HEN_FILE_EXTENSION
        )
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join([self.f0, self.f1, self.f2, "-o", out])
        assert np.any(
            [
                f"{self.f1} has a different MJDREF" in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(out)
        os.unlink(out)
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join([self.f0, self.f2, self.f3, "-o", out])
        assert np.any(
            [
                f"{self.f3} is from a different" in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(out)
        os.unlink(out)
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join([self.f0, self.f2, self.f4, "-o", out])
        assert np.any(
            [
                f"{self.f4} has no good events" in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(out)
        os.unlink(out)

    def test_merge_many_events(self):
        outfile = "joint_ev" + HEN_FILE_EXTENSION
        # Note that only 0 and 2 are valid
        with pytest.warns(UserWarning) as record:
            hen.read_events.main_join([self.f0, self.f2, self.f3])

        assert os.path.exists(outfile)

        data = load_events(outfile)
        assert hasattr(data, "gti")
        assert data.gti is not None
        allgtis = []
        # Note that only 0 and 2 are valid
        for evfile in [self.f0, self.f2]:
            ev = load_events(evfile)
            allgtis.append(ev.gti)
        allgtis = np.sort(np.concatenate(allgtis))
        assert np.allclose(data.gti, allgtis)
        os.unlink(outfile)


class TestReadEvents:
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.fits_fileA = os.path.join(cls.datadir, "monol_testA.evt")
        cls.fits_fileB = os.path.join(cls.datadir, "monol_testB.evt")
        cls.fits_file = os.path.join(cls.datadir, "monol_test_fake.evt")
        main(
            [
                "--deadtime",
                "1e-4",
                "-m",
                "XMM",
                "-i",
                "epn",
                "--ctrate",
                "2000",
                "--mjdref",
                "50814.0",
                "-o",
                cls.fits_file,
            ]
        )
        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileB = os.path.join(
            cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
        )

    def test_start(self):
        """Make any warnings in setup_class be dumped here."""
        pass

    def test_treat_event_file_nustar(self):
        treat_event_file(self.fits_fileA)
        new_filename = "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        assert os.path.exists(os.path.join(self.datadir, new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert "instr" in data
        assert "gti" in data
        assert "mjdref" in data
        assert np.isclose(data["mjdref"], ref_mjd(self.fits_fileA))

    def test_treat_event_file_xmm(self):
        treat_event_file(self.fits_file)
        new_filename = "monol_test_fake_xmm_epn_det01_ev" + HEN_FILE_EXTENSION
        assert os.path.exists(os.path.join(self.datadir, new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert "instr" in data
        assert "gti" in data
        assert "mjdref" in data
        assert "pi" in data and data["pi"].size > 0

    def test_treat_event_file_xmm_gtisplit(self):

        treat_event_file(self.fits_file, gti_split=True)
        new_filename = (
            "monol_test_fake_xmm_epn_det01_gti000_ev" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(os.path.join(self.datadir, new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert "instr" in data
        assert "gti" in data
        assert "mjdref" in data

    def test_treat_event_file_xmm_lensplit(self):

        treat_event_file(self.fits_file, length_split=100)
        new_filename = (
            "monol_test_fake_xmm_epn_det01_chunk000_ev" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(os.path.join(self.datadir, new_filename))
        data = load_data(os.path.join(self.datadir, new_filename))
        assert "instr" in data
        assert "gti" in data
        assert "mjdref" in data
        gtis = data["gti"]
        lengths = np.array([g1 - g0 for (g0, g1) in gtis])
        # add an epsilon for numerical error
        assert np.all(lengths <= 100 + 1e-7)

    def test_split_events(self):
        treat_event_file(self.fits_fileA)

        filea = os.path.join(
            self.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )

        files = hen.read_events.main_splitevents([filea, "-l", "50"])
        for f in files:
            assert os.path.exists(f)

    def test_split_events_bad_overlap_raises(self):
        treat_event_file(self.fits_fileA)

        filea = os.path.join(
            self.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )

        with pytest.raises(ValueError) as excinfo:
            hen.read_events.split_eventlist(filea, 10, overlap=1.5)
        assert "Overlap cannot be >=1. Exiting." in str(excinfo.value)

    def test_load_events(self):
        """Test event file reading."""
        command = "{}".format(self.fits_fileA)
        hen.read_events.main(command.split())
        ev = hen.io.load_events(self.ev_fileA)
        assert hasattr(ev, "header")
        assert hasattr(ev, "gti")

        main_readfile([self.ev_fileA])

    def test_load_events_with_2_cpus(self):
        """Test event file reading."""
        command = "{} {} --nproc 2".format(
            os.path.join(self.datadir, "monol_testB.evt"),
            os.path.join(self.datadir, "monol_testA_timezero.evt"),
        )
        hen.read_events.main(command.split())

    def test_load_events_split(self):
        """Test event file splitting."""
        command = "{0} -g --min-length 0".format(self.fits_fileB)
        hen.read_events.main(command.split())
        new_filename = os.path.join(
            self.datadir,
            "monol_testB_nustar_fpmb_gti000_ev" + HEN_FILE_EXTENSION,
        )
        assert os.path.exists(new_filename)
        command = "{0}".format(new_filename)
        hen.lcurve.main(command.split())
        new_filename = os.path.join(
            self.datadir,
            "monol_testB_nustar_fpmb_gti000_lc" + HEN_FILE_EXTENSION,
        )
        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.ev_fileB).gti[0]
        assert np.allclose(gti_to_test, lc.gti)

    def test_load_events_noclobber(self):
        """Test event file reading w. noclobber option."""
        with catch_warnings() as w:
            command = "{0} --noclobber".format(self.fits_fileB)
            hen.read_events.main(command.split())
        assert (
            str(w[0].message)
            .strip()
            .endswith("exists and using noclobber. Skipping")
        ), "Unexpected warning output"

    @classmethod
    def teardown_class(cls):
        for pattern in [
            "monol_*" + HEN_FILE_EXTENSION,
            "*phasetag*",
            "*fake*",
            "monol*.pdf",
        ]:
            files = glob.glob(os.path.join(cls.datadir, pattern))
            for file in files:
                os.unlink(file)



from stingray.events import EventList
from stingray.lightcurve import Lightcurve
import numpy as np
from hendrics.base import HAS_NUMBA
from ..ffa import ffa_search

from ..efsearch import fit
import pytest

# @pytest.mark.skipif('not HAS_NUMBA')
def test_ffa():
    period = 0.01
    pmin = 0.0095
    pmax = 0.0105
    dt = 10 ** int(np.log10(period)) / 256
    length = 1
    times = np.arange(0, length, dt)

    flux = 10 + np.cos(2 * np.pi * times / period)

    lc_cont = Lightcurve(times, flux, err_dist="gauss")

    ev = EventList()
    ev.simulate_times(lc_cont)
    lc = Lightcurve.make_lightcurve(ev.time, dt=dt, tstart=0, tseg=length)

    per, st = ffa_search(lc.counts, dt, pmin, pmax)
    #  fit_sinc wants frequencies, not periods
    model = fit(1 / per[::-1], st[::-1], 1 / period, obs_length=10)
    assert np.isclose(1 / model.mean, period, atol=1e-6)


# @pytest.mark.skipif('not HAS_NUMBA')
def test_ffa_large_intv():
    period = 0.01
    pmin = 0.002789345
    pmax = 0.0105
    dt = 10 ** int(np.log10(0.002789345)) / 20
    length = 5
    times = np.arange(0, length, dt)

    flux = 10 + np.cos(2 * np.pi * times / period)

    lc_cont = Lightcurve(times, flux, err_dist="gauss")

    ev = EventList()
    ev.simulate_times(lc_cont)
    lc = Lightcurve.make_lightcurve(ev.time, dt=dt, tstart=0, tseg=length)

    per, st = ffa_search(lc.counts, dt, pmin, pmax)
    #  fit_sinc wants frequencies, not periods
    model = fit(1 / per[::-1], st[::-1], 1 / period, obs_length=10)
    assert np.isclose(1 / model.mean, period, atol=1e-6)


# @pytest.mark.skipif('not HAS_NUMBA')


def test_ffa_vs_folding_search():
    import time
    from hendrics.efsearch import folding_search
    from stingray.events import EventList
    from stingray.lightcurve import Lightcurve

    period = 0.01
    pmin = 0.0095
    pmax = 0.0105
    dt = 10 ** int(np.log10(period)) / 256
    length = 1
    times = np.arange(0, length, dt)

    flux = 5 + 1 * np.cos(2 * np.pi * times / period)

    lc_cont = Lightcurve(times, flux, err_dist="gauss")

    ev = EventList()
    ev.simulate_times(lc_cont)
    lc = Lightcurve.make_lightcurve(ev.time, dt=dt, tstart=0, tseg=length)

    t0 = time.time()
    per, st = ffa_search(lc.counts, dt, pmin, pmax)
    t1 = time.time()
    print("FFA completed in {:.1e} s".format(t1 - t0))
    t1 = time.time()
    freqs, stats, _, _ = folding_search(
        ev, 1 / pmax, 1 / pmin, oversample=3, nbin=128
    )
    t2 = time.time()
    print("Standard search completed in {:.1e} s".format(t2 - t1))

    comparable_stats = np.array(
        [st[idx] for idx in np.searchsorted(per, 1 / freqs)]
    )
    assert (comparable_stats - stats + 127).std() < 127



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Quicklook plots."""

import warnings
import os
import copy
from collections.abc import Iterable
import numpy as np
from stingray.gti import create_gti_mask
from astropy.modeling.models import Const1D
from astropy.modeling import Model
from astropy.stats import poisson_conf_interval
from astropy import log
from astropy.table import Table

from .fold import fold_events, filter_energy
from .base import z2_n_detection_level
from .base import fold_detection_level
from .base import deorbit_events
from .io import load_events
from .io import load_data, get_file_type, load_pds
from .io import is_string, save_as_qdp, load_folding
from .io import HEN_FILE_EXTENSION
from .io import find_file_in_allowed_paths
from .base import _assign_value_if_none, find_peaks_in_image
from .base import pds_detection_level as detection_level


def _next_color(ax):
    try:
        return next(ax._get_lines.color_cycle)
    except Exception:
        return next(ax._get_lines.prop_cycler)["color"]


def _baseline_fun(x, a):
    """A constant function."""
    return a


def _value_or_none(dict_like, key):
    try:
        return dict_like[key]
    except KeyError:
        return None


def plot_generic(
    fnames,
    vars,
    errs=None,
    figname=None,
    xlog=None,
    ylog=None,
    output_data_file=None,
):
    """Generic plotting function."""
    import matplotlib.pyplot as plt

    if is_string(fnames):
        fnames = [fnames]
    figname = _assign_value_if_none(
        figname, "{0} vs {1}".format(vars[1], vars[0])
    )
    plt.figure(figname)
    ax = plt.gca()
    if xlog:
        ax.set_xscale("log", nonpositive="clip")
    if ylog:
        ax.set_yscale("log", nonpositive="clip")

    xlabel, ylabel = vars
    xlabel_err, ylabel_err = None, None
    if errs is not None:
        xlabel_err, ylabel_err = errs

    for i, fname in enumerate(fnames):
        data = load_data(fname)
        color = _next_color(ax)
        xdata = data[xlabel]
        ydata = data[ylabel]
        xdata_err = _value_or_none(data, xlabel_err)
        ydata_err = _value_or_none(data, ylabel_err)
        plt.errorbar(
            xdata,
            ydata,
            yerr=ydata_err,
            xerr=xdata_err,
            fmt="-",
            drawstyle="steps-mid",
            color=color,
            label=fname,
        )

        if output_data_file is not None:
            save_as_qdp(
                [xdata, ydata],
                errors=[xdata_err, ydata_err],
                filename=output_data_file,
                mode="a",
            )

    plt.xlabel(vars[0])
    plt.ylabel(vars[1])
    plt.legend()


def _get_const(models):
    """Get constant from Astropy model, list of models or compound model.

    Return None if no Const1D objects are in ``models``.
    Return the value of the first Const1D object found.

    Examples
    --------
    >>> from astropy.modeling.models import Const1D, Gaussian1D
    >>> model = Const1D(2) + Gaussian1D(1, 4, 5)
    >>> _get_const(model)
    2.0
    >>> _get_const(model[0])
    2.0
    >>> _get_const([model[0]])
    2.0
    >>> _get_const([[model]])
    2.0
    >>> _get_const(model[1])

    >>> _get_const(None)

    >>> _get_const(1)

    >>> _get_const('avdsfa')

    """

    if isinstance(models, Const1D):
        return models.amplitude.value

    if hasattr(models, "submodel_names"):
        for subm in models:
            if isinstance(subm, Const1D):
                return subm.amplitude.value

    if models is None:
        return None

    if (
        isinstance(models, Iterable)
        and not is_string(models)
        and len(models) != 0
    ):
        for model in models:
            return _get_const(model)

    return None


def plot_pds(
    fnames,
    figname=None,
    xlog=None,
    ylog=None,
    output_data_file=None,
    white_sub=False,
):
    """Plot a list of PDSs, or a single one."""

    from scipy.optimize import curve_fit
    import matplotlib.pyplot as plt

    if is_string(fnames):
        fnames = [fnames]

    figlabel = fnames[0]

    for i, fname in enumerate(fnames):
        pds_obj = load_pds(fname, nosub=True)
        if np.allclose(np.diff(pds_obj.freq), pds_obj.df):
            freq = pds_obj.freq
            xlog = _assign_value_if_none(xlog, False)
            ylog = _assign_value_if_none(ylog, False)
        else:
            flo = pds_obj.freq - pds_obj.df / 2
            fhi = pds_obj.freq + pds_obj.df / 2
            freq = (fhi + flo) / 2
            xlog = _assign_value_if_none(xlog, True)
            ylog = _assign_value_if_none(ylog, True)

        models = []
        if hasattr(pds_obj, "best_fits") and pds_obj.best_fits is not None:
            models = pds_obj.best_fits
        if isinstance(models, Model):
            models = [models]

        pds = pds_obj.power
        epds = pds_obj.power_err
        npds = pds_obj.m
        norm = pds_obj.norm

        lev = detection_level(
            epsilon=0.015, n_summed_spectra=npds, ntrial=pds.size
        )

        if norm == "rms":
            # we need the unnormalized power
            lev = lev / 2 * pds_obj.nphots
            lev, _ = pds_obj._normalize_crossspectrum(lev, pds_obj.fftlen)

        if xlog and ylog:
            plt.figure("PDS - Loglog " + figlabel)
        else:
            plt.figure("PDS " + figlabel)
        ax = plt.gca()
        color = _next_color(ax)

        if xlog:
            ax.set_xscale("log", nonpositive="clip")
        if ylog:
            ax.set_yscale("log", nonpositive="clip")

        level = lev  # Can be modified below
        y = pds[1:]
        yerr = epds[1:]

        if norm.lower() == "leahy" or (
            norm.lower() in ["rms", "frac"] and (not xlog or not ylog)
        ):
            plt.plot(
                freq[1:], y, drawstyle="steps-mid", color=color, label=fname
            )
            for i, func in enumerate(models):
                plt.plot(
                    freq,
                    func(freq),
                    label="Model {}".format(i + 1),
                    zorder=20,
                    color="k",
                )

        elif norm.lower() in ["rms", "frac"] and xlog and ylog:
            # TODO: Very rough! Use new machinery
            const = _get_const(models)
            if const is None:
                p, pcov = curve_fit(
                    _baseline_fun, freq, pds, p0=[2], sigma=epds
                )
                log.info("White noise level is {0}".format(p[0]))
                const = p[0]

            pds -= const
            level = lev - const

            y = pds[1:] * freq[1:]
            yerr = epds[1:] * freq[1:]
            plt.plot(
                freq[1:], y, drawstyle="steps-mid", color=color, label=fname
            )
            level *= freq
            for i, func in enumerate(models):
                const = _get_const(func)
                plt.plot(
                    freq,
                    freq * (func(freq) - const),
                    label="Model {}".format(i + 1),
                    zorder=20,
                    color="k",
                )

        if np.any(level < 0):
            continue
        if isinstance(level, Iterable):
            plt.plot(freq, level, ls="--", color=color)
        else:
            plt.axhline(level, ls="--", color=color)
        if output_data_file is not None:
            save_as_qdp(
                [freq[1:], y],
                errors=[None, yerr],
                filename=output_data_file,
                mode="a",
            )

    plt.xlabel("Frequency")
    if norm.lower() == "rms":
        plt.ylabel("(rms/mean)^2")
    elif norm.lower() == "leahy":
        plt.ylabel("Leahy power")

    plt.legend()

    if figname is not None:
        plt.savefig(figname)


def plot_cospectrum(
    fnames, figname=None, xlog=None, ylog=None, output_data_file=None
):
    """Plot the cospectra from a list of CPDSs, or a single one."""
    import matplotlib.pyplot as plt

    if is_string(fnames):
        fnames = [fnames]

    figlabel = fnames[0]

    for i, fname in enumerate(fnames):
        pds_obj = load_pds(fname, nosub=True)
        models = []
        if hasattr(pds_obj, "best_fits") and pds_obj.best_fits is not None:
            models = pds_obj.best_fits

        if np.allclose(np.diff(pds_obj.freq), pds_obj.df):
            freq = pds_obj.freq
            xlog = _assign_value_if_none(xlog, False)
            ylog = _assign_value_if_none(ylog, False)
        else:
            flo = pds_obj.freq - pds_obj.df / 2
            fhi = pds_obj.freq + pds_obj.df / 2
            freq = (fhi + flo) / 2
            xlog = _assign_value_if_none(xlog, True)
            ylog = _assign_value_if_none(ylog, True)

        cpds = pds_obj.power

        cospectrum = cpds.real
        if xlog and ylog:
            plt.figure("Cospectrum - Loglog " + figlabel)
        else:
            plt.figure("Cospectrum " + figlabel)
        ax = plt.gca()
        if xlog:
            ax.set_xscale("log", nonpositive="clip")
        if ylog:
            ax.set_yscale("log", nonpositive="clip")

        plt.xlabel("Frequency")
        if xlog and ylog:
            y = freq[1:] * cospectrum[1:]
            plt.plot(freq[1:], y, drawstyle="steps-mid", label=fname)
            for i, func in enumerate(models):
                plt.plot(
                    freq, freq * func(freq), label="Model {}".format(i + 1)
                )

            plt.ylabel("Cospectrum * Frequency")
        else:
            y = cospectrum[1:]
            plt.plot(
                freq[1:], cospectrum[1:], drawstyle="steps-mid", label=fname
            )

            plt.ylabel("Cospectrum")
            for i, func in enumerate(models):
                plt.plot(freq, func(freq), label="Model {}".format(i + 1))
        if output_data_file is not None:
            save_as_qdp([freq[1:], y], filename=output_data_file, mode="a")

    plt.legend()

    if figname is not None:
        plt.savefig(figname)


def plot_folding(
    fnames, figname=None, xlog=None, ylog=None, output_data_file=None
):
    from matplotlib import gridspec
    import matplotlib.pyplot as plt

    if is_string(fnames):
        fnames = [fnames]

    for fname in fnames:
        ef = load_folding(fname)

        if not hasattr(ef, "M") or ef.M is None:
            ef.M = 1
        label = "Stat"
        ntrial = ef.stat.size
        if hasattr(ef, "oversample") and ef.oversample is not None:
            ntrial /= ef.oversample
            ntrial = int(ntrial)
        if ef.kind == "Z2n":
            vmin = ef.N - 1
            vmax = z2_n_detection_level(
                epsilon=0.001,
                n=int(ef.N),
                ntrial=ntrial,
                n_summed_spectra=int(ef.M),
            )
            nbin = ef.N * 8
            label = "$" + f"Z^2_{ef.N}" + "$"
        else:
            vmin = ef.nbin
            vmax = fold_detection_level(
                nbin=int(ef.nbin), epsilon=0.001, ntrial=ntrial
            )
            nbin = ef.nbin

        n_cands = 5
        best_cands = find_peaks_in_image(ef.stat, n=n_cands)

        fddot = 0
        if hasattr(ef, "fddots") and ef.fddots is not None:
            fddot = ef.fddots

        print("Best candidates:")
        best_cand_table = Table(names=["mjd", "power", "f", "fdot", "fddot"])
        # I do that in reverse order, so that the final solution is also the
        # best one, for plotting the candidate f, fdot
        for i, idx in enumerate(best_cands[::-1]):
            if len(ef.stat.shape) > 1 and ef.stat.shape[0] > 1:
                allfreqs = ef.freq[idx[0], :]
                allfdots = ef.freq[:, idx[1]]
                allstats_f = ef.stat[idx[0], :]
                allstats_fdot = ef.stat[:, idx[1]]
                f, fdot = ef.freq[idx[0], idx[1]], ef.fdots[idx[0], idx[1]]
                max_stat = ef.stat[idx[0], idx[1]]
            elif len(ef.stat.shape) == 1:
                allfreqs = ef.freq
                allstats_f = ef.stat
                f = ef.freq[idx[0]]
                max_stat = ef.stat[idx[0]]
                fdot = 0
                allfdots = None
                allstats_fdot = None
            else:
                raise ValueError("Did not understand stats shape.")
            if max_stat < vmax:
                continue
            best_cand_table.add_row([ef.pepoch, max_stat, f, fdot, fddot])
            Table({"freq": allfreqs, "stat": allstats_f}).write(
                f'{fname.replace(HEN_FILE_EXTENSION, "")}'
                f"_cand_{n_cands - i - 1}_fdot{fdot}.csv",
                overwrite=True,
                format="ascii",
            )
            if allfdots is None:
                continue

            Table({"fdot": allfdots, "stat": allstats_fdot}).write(
                f'{fname.replace(HEN_FILE_EXTENSION, "")}'
                f"_cand_{n_cands - i - 1}_f{f}.dat",
                overwrite=True,
                format="ascii",
            )

        if len(best_cand_table) == 0:
            print(f"None.")
            if hasattr(ef, "upperlim") and ef.upperlim is not None:
                uplim = ef.upperlim
                print(f"(Upper limit for sinusoids: p. frac. < {uplim})")
        else:
            print(best_cand_table[::-1])
        best_cand_table.write(fname + "_best_cands.csv", overwrite=True)
        plt.figure(fname, figsize=(8, 8))

        if (
            hasattr(ef, "filename")
            and ef.filename is not None
            and os.path.exists(ef.filename)
        ):
            external_gs = gridspec.GridSpec(2, 1)
            search_gs_no = 1

            events = load_events(ef.filename)
            if ef.emin is not None or ef.emax is not None:
                events, elabel = filter_energy(events, ef.emin, ef.emax)

            if hasattr(ef, "parfile") and ef.parfile is not None:
                root = os.path.split(fname)[0]
                parfile = find_file_in_allowed_paths(ef.parfile, [".", root])
                if not parfile:
                    warnings.warn("{} does not exist".format(ef.parfile))
                else:
                    ef.parfile = parfile

                if parfile and os.path.exists(parfile):
                    events = deorbit_events(events, parfile)

            if hasattr(ef, "ref_time") and ef.ref_time is not None:
                ref_time = ef.ref_time
            elif hasattr(ef, "pepoch") and ef.pepoch is not None:
                ref_time = (ef.pepoch - events.mjdref) * 86400
            else:
                ref_time = (events.time[0] + events.time[-1]) / 2

            pepoch = ref_time / 86400 + events.mjdref

            phase, profile, profile_err = fold_events(
                copy.deepcopy(events.time),
                f,
                fdot,
                ref_time=ref_time,
                # gtis=copy.deepcopy(events.gti),
                expocorr=False,
                nbin=nbin,
            )
            ax = plt.subplot(external_gs[0])
            Table(
                {
                    "phase": np.concatenate((phase, phase + 1)),
                    "profile": np.concatenate((profile, profile)),
                    "err": np.concatenate((profile_err, profile_err)),
                }
            ).write(
                f'{fname.replace(HEN_FILE_EXTENSION, "")}_folded.csv',
                overwrite=True,
                format="ascii",
            )

            # print(df, dfdot)
            # # noinspection PyPackageRequirements
            # ax.text(0.1, 0.9, "Profile for F0={} Hz, F1={} Hz/s".format(
            #     round(f, -int(np.floor(np.log10(np.abs(df))))),
            #     round(fdot, -int(np.floor(np.log10(np.abs(dfdot)))))),
            #     horizontalalignment='left', verticalalignment='center',
            #     transform=ax.transAxes)
            ax.plot(
                np.concatenate((phase, phase + 1)),
                np.concatenate((profile, profile)),
                drawstyle="steps-mid",
            )

            mean = np.mean(profile)

            low, high = poisson_conf_interval(
                mean, interval="frequentist-confidence", sigma=1
            )

            ax.axhline(mean)
            ax.fill_between(
                [0, 2],
                [low, low],
                [high, high],
                label=r"1-$\sigma c.l.$",
                alpha=0.5,
            )
            low, high = poisson_conf_interval(
                mean, interval="frequentist-confidence", sigma=3
            )
            ax.fill_between(
                [0, 2],
                [low, low],
                [high, high],
                label=r"3-$\sigma c.l.$",
                alpha=0.5,
            )
            ax.set_xlabel("Phase")
            ax.set_ylabel("Counts")
            ax.set_xlim([0, 2])
            ax.legend(loc=4)
            phascommand = (
                "HENphaseogram -f {} "
                "--fdot {} {} -n {} --norm to1".format(
                    f, fdot, ef.filename, nbin
                )
            )
            if ef.parfile and os.path.exists(ef.parfile):
                phascommand += " --deorbit-par {}".format(parfile)
            if hasattr(ef, "emin") and ef.emin is not None:
                phascommand += " --emin {}".format(ef.emin)
            if hasattr(ef, "emin") and ef.emin is not None:
                phascommand += " --emax {}".format(ef.emax)

            if hasattr(events, "mjdref") and events.mjdref is not None:
                phascommand += " --pepoch {}".format(pepoch)

            log.info(
                "To see the detailed phaseogram, " "run {}".format(phascommand)
            )

        elif not os.path.exists(ef.filename):
            warnings.warn(ef.filename + " does not exist")
            external_gs = gridspec.GridSpec(1, 1)
            search_gs_no = 0
        else:
            external_gs = gridspec.GridSpec(1, 1)
            search_gs_no = 0

        if len(ef.stat.shape) > 1 and ef.stat.shape[0] > 1:
            gs = gridspec.GridSpecFromSubplotSpec(
                2,
                2,
                height_ratios=(1, 3),
                width_ratios=(3, 1),
                hspace=0,
                wspace=0,
                subplot_spec=external_gs[search_gs_no],
            )

            axf = plt.subplot(gs[0, 0])
            axfdot = plt.subplot(gs[1, 1])
            if vmax is not None:
                axf.axhline(vmax, ls="--", label=r"99.9\% c.l.")
                axfdot.axvline(vmax)
            axffdot = plt.subplot(gs[1, 0], sharex=axf, sharey=axfdot)
            axffdot.pcolormesh(
                ef.freq,
                np.asarray(ef.fdots),
                ef.stat,
                vmin=vmin,
                vmax=vmax,
                shading="nearest",
            )
            maximum_idx = 0
            maximum = 0

            for ix in range(ef.stat.shape[0]):
                if ef.stat.shape[0] < 100:
                    axf.plot(
                        ef.freq[ix, :],
                        ef.stat[ix, :],
                        alpha=0.5,
                        lw=0.2,
                        color="k",
                    )
                if np.max(ef.stat[ix, :]) > maximum:
                    maximum = np.max(ef.stat[ix, :])
                    maximum_idx = ix
            if vmax is not None and maximum_idx > 0:
                axf.plot(
                    ef.freq[maximum_idx, :],
                    ef.stat[maximum_idx, :],
                    lw=1,
                    color="k",
                )
            maximum_idx = -1
            maximum = 0
            for iy in range(ef.stat.shape[1]):
                if ef.stat.shape[1] < 100:
                    axfdot.plot(
                        ef.stat[:, iy],
                        np.asarray(ef.fdots)[:, iy],
                        alpha=0.5,
                        lw=0.2,
                        color="k",
                    )
                if np.max(ef.stat[:, iy]) > maximum:
                    maximum = np.max(ef.stat[:, iy])
                    maximum_idx = iy
            if vmax is not None and maximum_idx > 0:
                axfdot.plot(
                    ef.stat[:, maximum_idx],
                    np.asarray(ef.fdots)[:, maximum_idx],
                    lw=1,
                    color="k",
                )
            axf.set_ylabel(label)
            axfdot.set_xlabel(label)

            # plt.colorbar()
            axffdot.set_xlabel("Frequency (Hz)")
            axffdot.set_ylabel("Fdot (Hz/s)")
            axffdot.set_xlim([np.min(ef.freq), np.max(ef.freq)])
            axffdot.set_ylim([np.min(ef.fdots), np.max(ef.fdots)])
            axffdot.axvline(f, ls="--", color="white")
            axffdot.axhline(fdot, ls="--", color="white")
            axf.legend(loc=4)
        else:
            axf = plt.subplot(external_gs[search_gs_no])
            axf.plot(ef.freq, ef.stat, drawstyle="steps-mid", label=fname)
            axf.set_xlabel("Frequency (Hz)")
            axf.set_ylabel(ef.kind + " stat")
            axf.legend(loc=4)

        if (
            hasattr(ef, "best_fits")
            and ef.best_fits is not None
            and not len(ef.stat.shape) > 1
        ):

            for f in ef.best_fits:
                xs = np.linspace(
                    np.min(ef.freq), np.max(ef.freq), len(ef.freq) * 2
                )
                plt.plot(xs, f(xs))

        if output_data_file is not None:
            fdots = ef.fdots
            if not isinstance(fdots, Iterable) or len(fdots) == 1:
                fdots = fdots + np.zeros_like(ef.freq.flatten())
            # print(fdots.shape, ef.freq.shape, ef.stat.shape)
            out = [ef.freq.flatten(), fdots.flatten(), ef.stat.flatten()]
            out_err = [None, None, None]

            if (
                hasattr(ef, "best_fits")
                and ef.best_fits is not None
                and not len(ef.stat.shape) > 1
            ):
                for f in ef.best_fits:
                    out.append(f(ef.freq.flatten()))
                    out_err.append(None)

            save_as_qdp(out, out_err, filename=output_data_file, mode="a")

    ax = plt.gca()
    if xlog:
        ax.set_xscale("log", nonpositive="clip")
    if ylog:
        ax.set_yscale("log", nonpositive="clip")
    try:
        plt.tight_layout()
    except Exception:  # pragma: no cover
        pass

    if figname is not None:
        plt.savefig(figname)


def plot_color(
    file0, file1, xlog=None, ylog=None, figname=None, output_data_file=None
):
    import matplotlib.pyplot as plt

    type0, lc0 = get_file_type(file0)
    type1, lc1 = get_file_type(file1)
    xlabel, ylabel = "Count rate", "Count rate"
    if type0 == "color":
        xlabel = "{3}-{2}/{1}-{0}".format(*lc0.e_intervals)
    if type1 == "color":
        ylabel = "{3}-{2}/{1}-{0}".format(*lc1.e_intervals)
    plt.errorbar(
        lc0.counts,
        lc1.counts,
        xerr=lc0.counts_err,
        yerr=lc1.counts_err,
        fmt="o",
        color="k",
        alpha=0.5,
    )
    plt.scatter(lc0.counts, lc1.counts, zorder=10)

    if output_data_file is not None:
        save_as_qdp(
            [lc0.counts, lc1.counts],
            errors=[lc0.counts_err, lc1.counts_err],
            filename=output_data_file,
            mode="a",
        )

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    ax = plt.gca()
    if xlog:
        ax.set_xscale("log", nonpositive="clip")
    if ylog:
        ax.set_yscale("log", nonpositive="clip")
    if figname is not None:
        plt.savefig(figname)


def plot_lc(
    lcfiles,
    figname=None,
    fromstart=False,
    xlog=None,
    ylog=None,
    output_data_file=None,
):
    """Plot a list of light curve files, or a single one."""
    import matplotlib.pyplot as plt

    if is_string(lcfiles):
        lcfiles = [lcfiles]

    figlabel = lcfiles[0]

    plt.figure("LC " + figlabel)
    for lcfile in lcfiles:
        log.info("Loading %s..." % lcfile)
        lcdata = load_data(lcfile)

        time = lcdata["time"]
        lc = lcdata["counts"]
        gti = lcdata["gti"]
        instr = lcdata["instr"]

        if fromstart:
            time -= lcdata["Tstart"]
            gti -= lcdata["Tstart"]

        if instr == "PCA":
            # If RXTE, plot per PCU count rate
            npcus = lcdata["nPCUs"]
            lc /= npcus

        bti = list(zip(gti[:-1, 1], gti[1:, 0]))

        for g in bti:
            plt.axvspan(g[0], g[1], color="red", alpha=0.5)

        good = create_gti_mask(time, gti)
        plt.plot(time, lc, drawstyle="steps-mid", color="grey")
        plt.plot(time[good], lc[good], drawstyle="steps-mid", label=lcfile)
        if "base" in lcdata:
            plt.plot(time, lcdata["base"], color="r")

        if output_data_file is not None:
            outqdpdata = [time[good], lc[good]]
            if "base" in lcdata:
                outqdpdata.append(lcdata["base"][good])
            save_as_qdp(outqdpdata, filename=output_data_file, mode="a")

    plt.xlabel("Time (s)")
    if instr == "PCA":
        plt.ylabel("light curve (Ct/bin/PCU)")
    else:
        plt.ylabel("light curve (Ct/bin)")

    plt.legend()
    if figname is not None:
        plt.savefig(figname)


def main(args=None):
    """Main function called by the `HENplot` command line script."""
    import argparse
    from .base import check_negative_numbers_in_args

    description = (
        "Plot the content of HENDRICS light curves and frequency spectra"
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "--noplot",
        help="Only create images, do not plot",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--CCD",
        help="This is a color-color diagram. In this case, the"
        " list of files is expected to be given as "
        "soft0.nc, hard0.nc, soft1.nc, hard1.nc, ...",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--HID",
        help="This is a hardness-intensity diagram. In this "
        "case, the list of files is expected to be given "
        "as color0.nc, intensity0.nc, color1.nc, "
        "intensity1.nc, ...",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--figname", help="Figure name", default=None, type=str
    )
    parser.add_argument(
        "-o",
        "--outfile",
        help="Output data file in QDP format",
        default=None,
        type=str,
    )
    parser.add_argument(
        "--xlog",
        help="Use logarithmic X axis",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--ylog",
        help="Use logarithmic Y axis",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--xlin", help="Use linear X axis", default=None, action="store_true"
    )
    parser.add_argument(
        "--ylin", help="Use linear Y axis", default=None, action="store_true"
    )
    parser.add_argument(
        "--fromstart",
        help="Times are measured from the start of the "
        "observation (only relevant for light curves)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--axes",
        nargs=2,
        type=str,
        help="Plot two variables contained in the file",
        default=None,
    )

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)
    if args.noplot and args.figname is None:
        args.figname = args.files[0].replace(HEN_FILE_EXTENSION, ".png")
        import matplotlib

        matplotlib.use("Agg")
    if args.xlin is not None:
        args.xlog = False
    if args.ylin is not None:
        args.ylog = False

    if args.CCD or args.HID:
        args.files = zip(args.files[:-1:2], args.files[1::2])

    for fname in args.files:

        if args.CCD or args.HID:
            plot_color(
                fname[0],
                fname[1],
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )
            continue
        ftype, contents = get_file_type(fname)
        if args.axes is not None:
            plot_generic(
                fname,
                args.axes,
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )
            continue
        if ftype == "lc":
            plot_lc(
                fname,
                fromstart=args.fromstart,
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )
        elif ftype == "folding":
            plot_folding(
                fname,
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )
        elif ftype[-4:] == "cpds":
            plot_cospectrum(
                fname,
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )
        elif ftype[-3:] == "pds":
            plot_pds(
                fname,
                xlog=args.xlog,
                ylog=args.ylog,
                figname=args.figname,
                output_data_file=args.outfile,
            )

    if not args.noplot:
        import matplotlib.pyplot as plt

        plt.show()



import os
import copy
import glob
from collections.abc import Iterable

import pytest
from stingray.lightcurve import Lightcurve
from stingray.events import EventList
import numpy as np
from hendrics.io import (
    save_events,
    HEN_FILE_EXTENSION,
    load_folding,
    load_events,
    get_file_type,
    save_lcurve,
)
from hendrics.efsearch import main_efsearch, main_zsearch
from hendrics.efsearch import main_accelsearch, main_z2vspf
from hendrics.efsearch import decide_binary_parameters, folding_orbital_search
from hendrics.fold import main_fold, main_deorbit, std_fold_fit_func
from hendrics.fold import fit_profile_with_sinusoids, get_TOAs_from_events
from hendrics.plot import plot_folding
from hendrics.tests import _dummy_par
from hendrics.base import hen_root
from astropy.tests.helper import remote_data

try:
    import pandas as pd

    HAS_PD = True
except ImportError:
    HAS_PD = False

from hendrics.fold import HAS_PINT
from hendrics.efsearch import HAS_IMAGEIO


class TestEFsearch:
    def setup_class(cls):
        cls.pulse_period = 0.101
        cls.pulse_frequency = 1 / cls.pulse_period
        cls.tstart = 0
        cls.tend = 25.25
        cls.tseg = cls.tend - cls.tstart
        cls.dt = 0.00606
        cls.times = np.arange(cls.tstart, cls.tend, cls.dt) + cls.dt / 2
        cls.counts = 200 + 40 * np.cos(
            2 * np.pi * cls.times * cls.pulse_frequency
        )
        cls.mjdref = 56000

        lc = Lightcurve(
            cls.times, cls.counts, gti=[[cls.tstart, cls.tend]], dt=cls.dt
        )
        cls.lcfile = "lcurve" + HEN_FILE_EXTENSION
        save_lcurve(lc, cls.lcfile)
        events = EventList()
        events.mjdref = cls.mjdref
        events.simulate_times(lc)
        events.mission = "nusboh"
        cls.event_times = events.time
        cls.dum_noe = "events_noe" + HEN_FILE_EXTENSION
        save_events(events, cls.dum_noe)
        events.pi = np.random.uniform(0, 1000, len(events.time))
        cls.dum_pi = "events_pi" + HEN_FILE_EXTENSION
        save_events(events, cls.dum_pi)
        events.energy = np.random.uniform(3, 79, len(events.time))
        cls.dum = "events" + HEN_FILE_EXTENSION
        cls.dum_scramble = "events_scramble" + HEN_FILE_EXTENSION
        save_events(events, cls.dum)
        events_scramble = copy.deepcopy(events)
        events_scramble.time = np.sort(
            np.random.uniform(cls.tstart, cls.tend, events.time.size)
        )
        save_events(events_scramble, cls.dum_scramble)
        cls.par = "bububububu.par"
        _dummy_par(cls.par)
        events = EventList(time=np.sort(np.random.uniform(0, 100, 200)))
        cls.empty = "empty_ev" + HEN_FILE_EXTENSION
        save_events(events, cls.empty)

    def test_get_TOAs(self):
        events = load_events(self.dum)
        toas, toaerrs = get_TOAs_from_events(
            events.time,
            self.tseg,
            self.pulse_frequency,
            gti=events.gti,
            nbin=32,
            mjdref=events.mjdref,
            template=None,
        )

        possible_toas = (
            events.mjdref + np.arange(-1, 3) * self.pulse_period / 86400
        )
        closest = possible_toas[np.argmin(np.abs(possible_toas - toas[0]))]

        delta_toa_s = (toas[0] - closest) * 86400
        toa_err_s = toaerrs[0] / 1e6

        assert np.abs(delta_toa_s) < toa_err_s * 4

    def test_get_TOAs_template(self):
        nbin = 32
        phases = np.arange(0, 1, 1 / nbin)
        template = np.cos(2 * np.pi * phases)
        events = load_events(self.dum)
        toas, toaerrs = get_TOAs_from_events(
            events.time,
            self.tseg,
            self.pulse_frequency,
            gti=events.gti,
            mjdref=events.mjdref,
            template=template,
            nbin=nbin,
        )
        possible_toas = (
            events.mjdref + np.arange(2) * self.pulse_period / 86400
        )
        closest = possible_toas[np.argmin(np.abs(possible_toas - toas[0]))]

        assert (toas[0] - closest) < toaerrs[0] / 86400000000

    def test_fit_profile_with_sinusoids(self):
        nbin = 32
        phases = np.arange(0, 1, 1 / nbin)
        prof_smooth = np.cos(2 * np.pi * phases) + 0.5 * np.cos(
            4 * np.pi * (phases + 0.5)
        )
        prof_smooth = (prof_smooth + 5) * 64
        prof = np.random.poisson(prof_smooth)
        baseline = np.mean(prof)
        proferr = np.sqrt(baseline)
        fit_pars_save, success_save, chisq_save = fit_profile_with_sinusoids(
            prof, proferr, debug=True, baseline=True
        )
        assert np.allclose(
            std_fold_fit_func(fit_pars_save, phases),
            prof_smooth,
            atol=3 * proferr,
        )

    def test_fold(self):
        evfile = self.dum
        evfile_noe = self.dum_noe
        evfile_pi = self.dum_pi

        main_fold(
            [
                evfile,
                "-f",
                str(self.pulse_frequency),
                "-n",
                "64",
                "--test",
                "--norm",
                "ratios",
            ]
        )
        outfile = "Energyprofile_ratios.png"
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold(
            [
                evfile_noe,
                "-f",
                str(self.pulse_frequency),
                "-n",
                "64",
                "--test",
                "--norm",
                "blablabla",
            ]
        )
        outfile = "Energyprofile.png"
        assert os.path.exists(outfile)
        os.unlink(outfile)
        main_fold(
            [
                evfile_pi,
                "-f",
                str(self.pulse_frequency),
                "-n",
                "64",
                "--test",
                "--norm",
                "to1",
            ]
        )
        outfile = "Energyprofile_to1.png"
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_fold_invalid(self):
        evfile = self.dum

        with pytest.raises(ValueError) as excinfo:
            main_fold(
                [
                    evfile,
                    "-f",
                    str(self.pulse_frequency),
                    "-n",
                    "64",
                    "--test",
                    "--norm",
                    "ratios",
                    "--pepoch",
                    str(self.mjdref),
                    "--tref",
                    "0",
                ]
            )
        assert "Only specify one between " in str(excinfo.value)

    def test_efsearch(self):
        evfile = self.dum
        main_efsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--emin",
                "3",
                "--emax",
                "79",
                "--fit-candidates",
            ]
        )
        outfile = "events_EF_3-79keV_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        ftype, efperiod = get_file_type(outfile)
        assert ftype == "folding"
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        os.unlink(outfile)

    def test_efsearch_from_lc(self):
        evfile = self.lcfile

        main_efsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fit-candidates",
            ]
        )
        outfile = "lcurve_EF_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )

    def test_zsearch(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--emin",
                "3",
                "--emax",
                "79",
                "--fit-candidates",
                "--fit-frequency",
                str(self.pulse_frequency),
                "--dynstep",
                "5",
            ]
        )
        outfile = "events_Z22_3-79keV_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        # Defaults to 2 harmonics
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_from_lc(self):
        evfile = self.lcfile

        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fit-candidates",
                "--fit-frequency",
                str(self.pulse_frequency),
                "--dynstep",
                "5",
            ]
        )
        outfile = "lcurve_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)
        efperiod = load_folding(outfile)
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        # Defaults to 2 harmonics
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fdots(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "32",
                "--fdotmin",
                " -0.01",
                "--fdotmax",
                "0.01",
                "--fit-candidates",
                "--fit-frequency",
                str(self.pulse_frequency),
            ]
        )
        outfile = "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file="bla.qdp")
        efperiod = load_folding(outfile)
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        # Defaults to 2 harmonics
        assert len(efperiod.fdots) > 1
        assert efperiod.N == 2
        os.unlink(outfile)

    @pytest.mark.skipif("not HAS_IMAGEIO")
    def test_transient(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "32",
                "--fdotmin",
                " -0.1",
                "--fdotmax",
                "0.1",
                "--transient",
                "--n-transient-intervals",
                "16",
            ]
        )
        outfile = "events_transient.gif"
        assert os.path.exists(outfile)
        os.unlink(outfile)

    @pytest.mark.skipif("HAS_IMAGEIO")
    def test_transient_warn_if_no_imageio(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            main_zsearch(
                [
                    evfile,
                    "-f",
                    "9.85",
                    "-F",
                    "9.95",
                    "-n",
                    "64",
                    "--fdotmin",
                    " -0.1",
                    "--fdotmax",
                    "0.1",
                    "--transient",
                ]
            )
        assert np.any(["imageio needed" in r.message.args[0] for r in record])

    def test_zsearch_print_upperlim(self):
        evfile = self.empty

        outfile = main_zsearch(
            [evfile, "-f", "4", "-F", "6", "-N", "1", "--fast"]
        )[0]
        plot_folding([outfile], ylog=True, output_data_file="bla.qdp")

        # assert "Upper limit for sinusoids:" in caplog.text
        os.unlink(outfile)

    def test_zsearch_fast(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fast",
                "--find-candidates",
                "--oversample",
                "4",
            ]
        )
        outfile = "events_Z22_9.85-9.95Hz_fast" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file="bla.qdp")
        efperiod = load_folding(outfile)

        assert len(efperiod.fdots) > 1
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fast_nofdot(self):
        evfile = self.dum
        outfiles = main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fast",
                "--fdotmin",
                0,
                "--fdotmax",
                "0",
            ]
        )
        outfile = outfiles[0]
        assert os.path.exists(outfile)
        efperiod = load_folding(outfile)

        assert (
            not isinstance(efperiod.fdots, Iterable)
            or len(efperiod.fdots) <= 1
        )
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fast_nbin_small_warns(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            _ = main_zsearch(
                [
                    evfile,
                    "-f",
                    "9.85",
                    "-F",
                    "9.95",
                    "-n",
                    "2",
                    "--fast",
                ]
            )
        assert np.any(
            [
                "The number of bins is too small" in r.message.args[0]
                for r in record
            ]
        )

    def test_zsearch_fdots_fast(self):
        evfile = self.dum
        outfiles = main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fast",
                "--mean-fdot",
                "1e-10",
            ]
        )

        outfile = outfiles[0]
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file="bla.qdp")
        efperiod = load_folding(outfile)

        assert len(efperiod.fdots) > 1
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fddots_fast(self):
        evfile = self.dum
        outfiles = main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fast",
                "--mean-fdot",
                "0",
                "--mean-fddot",
                "1e-13",
            ]
        )

        outfile = outfiles[0]
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file="bla.qdp")
        efperiod = load_folding(outfile)

        assert len(efperiod.fdots) > 1
        assert efperiod.N == 2
        os.unlink(outfile)

    def test_zsearch_fdots_ffa(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            main_zsearch(
                [
                    evfile,
                    "-f",
                    "9.89",
                    "-F",
                    "9.92",
                    "-n",
                    "32",
                    "--ffa",
                    "--find-candidates",
                ]
            )
        assert np.any(
            [
                "Folding Algorithm functionality" in r.message.args[0]
                for r in record
            ]
        )
        outfile = "events_Z22_9.89-9.92Hz_ffa" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True, output_data_file="bla_ffa.qdp")
        efperiod = load_folding(outfile)

        assert efperiod.N == 2
        assert np.isclose(
            efperiod.peaks[0], self.pulse_frequency, atol=1 / 25.25
        )
        os.unlink(outfile)

    def test_fold_fast_fails(self):
        evfile = self.dum

        with pytest.raises(ValueError) as excinfo:
            main_efsearch(
                [evfile, "-f", "9.85", "-F", "9.95", "-n", "64", "--fast"]
            )
        assert "The fast option is only available for z " in str(excinfo.value)

    def test_zsearch_fdots_fast_transient(self):
        evfile = self.dum
        main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--fast",
                "--transient",
                "--n-transient-intervals",
                "16",
            ]
        )

    @pytest.mark.skipif("not HAS_PD")
    def test_orbital(self):
        import pandas as pd

        events = load_events(self.dum)
        csv_file = decide_binary_parameters(
            137430,
            [0.03, 0.035],
            [2.0 * 86400, 2.5 * 86400],
            [0.0, 1.0],
            fdot_range=[0, 5e-10],
            reset=False,
            NMAX=10,
        )
        table = pd.read_csv(csv_file)
        assert len(table) == 10
        folding_orbital_search(
            events, csv_file, chunksize=10, outfile="out.csv"
        )
        table = pd.read_csv("out.csv")
        assert len(table) == 10
        assert np.all(table["done"])
        os.unlink(csv_file)
        os.unlink("out.csv")

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_efsearch_deorbit(self):
        evfile = self.dum

        ip = main_zsearch(
            [
                evfile,
                "-f",
                "9.85",
                "-F",
                "9.95",
                "-n",
                "64",
                "--deorbit-par",
                self.par,
            ]
        )

        outfile = "events_Z22_9.85-9.95Hz" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        plot_folding([outfile], ylog=True)

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_fold_deorbit(self):
        evfile = self.dum

        main_fold(
            [
                evfile,
                "-f",
                str(self.pulse_frequency),
                "-n",
                "64",
                "--test",
                "--norm",
                "ratios",
                "--deorbit-par",
                self.par,
                "--pepoch",
                str(self.mjdref),
            ]
        )
        outfile = "Energyprofile_ratios.png"
        assert os.path.exists(outfile)
        os.unlink(outfile)

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_deorbit(self):
        evfile = self.dum
        main_deorbit([evfile, "--deorbit-par", self.par])
        outfile = hen_root(evfile) + "_deorb" + HEN_FILE_EXTENSION
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_efsearch_deorbit_invalid(self):
        evfile = self.dum
        with pytest.raises(FileNotFoundError) as excinfo:
            ip = main_efsearch(
                [
                    evfile,
                    "-f",
                    "9.85",
                    "-F",
                    "9.95",
                    "-n",
                    "64",
                    "--deorbit-par",
                    "nonexistent.par",
                ]
            )
        assert "Parameter file" in str(excinfo.value)

    def test_accelsearch(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            outfile = main_accelsearch(
                [
                    evfile,
                    "--fmin",
                    "1",
                    "--fmax",
                    "10",
                    "--zmax",
                    "1",
                    "--delta-z",
                    "0.5",
                ]
            )
        assert np.any(
            [
                "The accelsearch functionality is " in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_accelsearch_nodetections(self):
        evfile = self.dum_scramble
        with pytest.warns(UserWarning) as record:
            outfile = main_accelsearch(
                [evfile, "--fmin", "1", "--fmax", "1.1", "--zmax", "1"]
            )
        assert np.any(
            [
                "The accelsearch functionality is " in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_accelsearch_energy_and_freq_filt(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            outfile = main_accelsearch(
                [
                    evfile,
                    "--emin",
                    "3",
                    "--emax",
                    "80",
                    "--fmin",
                    "0.1",
                    "--fmax",
                    "1",
                    "--zmax",
                    "5",
                ]
            )
        assert np.any(
            [
                "The accelsearch functionality is " in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_accelsearch_pad(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            outfile = main_accelsearch(
                [evfile, "--pad-to-double", "--zmax", "1"]
            )
        assert np.any(
            [
                "The accelsearch functionality is " in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_accelsearch_interbin(self):
        evfile = self.dum
        with pytest.warns(UserWarning) as record:
            outfile = main_accelsearch([evfile, "--interbin", "--zmax", "1"])
        assert np.any(
            [
                "The accelsearch functionality is " in r.message.args[0]
                for r in record
            ]
        )
        assert os.path.exists(outfile)
        os.unlink(outfile)

    def test_z2vspf(self):
        evfile = self.dum
        ip = main_z2vspf([evfile, "--show-z-values", "30", "--ntrial", "10"])

    @classmethod
    def teardown_class(cls):
        os.unlink(cls.empty)
        os.unlink(cls.dum_scramble)
        os.unlink(cls.dum_noe)
        os.unlink(cls.dum_pi)
        os.unlink(cls.par)
        for fname in glob.glob("*cand*.csv"):
            os.unlink(fname)
        for fname in glob.glob("*cand*.dat"):
            os.unlink(fname)



import os
import pytest
from astropy.tests.helper import remote_data
import numpy as np
from hendrics.base import deorbit_events
from stingray.events import EventList
from hendrics.tests import _dummy_par
from hendrics.fold import HAS_PINT


def test_deorbit_badpar():
    ev = np.asarray(1)
    with pytest.warns(UserWarning) as record:
        ev_deor = deorbit_events(ev, None)
    assert np.any(
        ["No parameter file specified" in r.message.args[0] for r in record]
    )
    assert ev_deor == ev


def test_deorbit_non_existing_par():
    ev = np.asarray(1)
    with pytest.raises(FileNotFoundError) as excinfo:
        ev_deor = deorbit_events(ev, "warjladsfjqpeifjsdk.par")
    assert "Parameter file warjladsfjqpeifjsdk.par does not exist" in str(
        excinfo.value
    )


@remote_data
@pytest.mark.skipif("not HAS_PINT")
def test_deorbit_bad_mjdref():
    from hendrics.base import deorbit_events

    ev = EventList(np.arange(100), gti=np.asarray([[0, 2]]))
    ev.mjdref = 2
    par = _dummy_par("bububu.par")
    with pytest.raises(ValueError) as excinfo:
        _ = deorbit_events(ev, par)
    assert "MJDREF is very low (<01-01-1950), " in str(excinfo.value)
    os.remove("bububu.par")


@remote_data
@pytest.mark.skipif("not HAS_PINT")
def test_deorbit_inverse():
    from hendrics.base import deorbit_events

    ev = EventList(
        np.sort(np.random.uniform(0, 1000, 10)),
        gti=np.asarray([[0, 1000]]),
        mjdref=55000,
    )
    par = _dummy_par("bububu.par", pb=1.0, a1=30)
    ev2 = deorbit_events(ev, par)
    ev3 = deorbit_events(ev, par, invert=True)
    assert np.allclose(ev.time - ev2.time, -(ev.time - ev3.time), atol=1e-6)
    os.remove("bububu.par")


@remote_data
@pytest.mark.skipif("not HAS_PINT")
def test_deorbit_run():
    from hendrics.base import deorbit_events

    ev = EventList(np.arange(0, 210000, 1000), gti=np.asarray([[0.0, 210000]]))

    ev.mjdref = 56000.0
    ev.ephem = "de200"
    par = _dummy_par("bububu.par")
    with pytest.warns(UserWarning) as record:
        _ = deorbit_events(ev, par)
    assert np.any(
        ["The observation is very long." in r.message.args[0] for r in record]
    )
    os.remove("bububu.par")



# -*- coding: utf-8 -*-
"""
@author: marta
"""

import warnings
from astropy import log
from astropy.logger import AstropyUserWarning
import numpy as np
from stingray.varenergyspectrum import (
    RmsEnergySpectrum,
    _decode_energy_specification,
)
from stingray.varenergyspectrum import LagEnergySpectrum

# from stingray.covariancespectrum import AveragedCovariancespectrum

from .base import hen_root, interpret_bintime
from .io import load_events
from .io import save_as_qdp


def main(args=None):
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = "Calculates variability-energy spectra"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-f",
        "--freq-interval",
        nargs=2,
        type=float,
        default=[0.0, 100],
        help="Frequence interval",
    )
    parser.add_argument(
        "--energy-values",
        nargs=4,
        type=str,
        default="0.3 12 5 lin".split(" "),
        help="Choose Emin, Emax, number of intervals,"
        "interval spacing, lin or log",
    )
    parser.add_argument(
        "--segment-size",
        type=float,
        default=512,
        help="Length of the light curve intervals to be " "averaged",
    )
    parser.add_argument(
        "--ref-band",
        nargs=2,
        type=float,
        default=None,
        help="Reference band when relevant",
    )
    parser.add_argument(
        "--rms", default=False, action="store_true", help="Calculate rms"
    )
    parser.add_argument(
        "--covariance",
        default=False,
        action="store_true",
        help="Calculate covariance spectrum",
    )
    parser.add_argument(
        "--use-pi",
        default=False,
        action="store_true",
        help="Energy intervals are specified as PI channels",
    )
    parser.add_argument(
        "--cross-instr",
        default=False,
        action="store_true",
        help="Use data files in pairs, for example with the"
        "reference band from one and the subbands from "
        "the  other (useful in NuSTAR and "
        "multiple-detector missions)",
    )
    parser.add_argument(
        "--lag",
        default=False,
        action="store_true",
        help="Calculate lag-energy",
    )

    _add_default_args(parser, ["bintime", "loglevel", "debug"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)
    args.bintime = np.longdouble(interpret_bintime(args.bintime))

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENvarenergy.log"):
        filelist = []
        energy_spec = (
            float(args.energy_values[0]),
            float(args.energy_values[1]),
            int(args.energy_values[2]),
            args.energy_values[3],
        )

        from .io import sort_files

        if args.cross_instr:
            log.info("Sorting file list")
            sorted_files = sort_files(args.files)

            warnings.warn(
                "Beware! For cpds and derivatives, I assume that the "
                "files are from only two instruments and in pairs "
                "(even in random order)"
            )

            instrs = list(sorted_files.keys())

            files1 = sorted_files[instrs[0]]
            files2 = sorted_files[instrs[1]]
        else:
            files1 = args.files
            files2 = args.files

        for fnames in zip(files1, files2):
            fname = fnames[0]
            fname2 = fnames[1]

            events = load_events(fname)
            events2 = load_events(fname2)
            if not args.use_pi and (
                events.energy is None or events2.energy is None
            ):
                raise ValueError(
                    "If --use-pi is not specified, event lists must "
                    "be calibrated! Please use HENcalibrate."
                )

            if args.rms:
                rms = RmsEnergySpectrum(
                    events,
                    args.freq_interval,
                    energy_spec,
                    segment_size=args.segment_size,
                    bin_time=args.bintime,
                    events2=events2,
                    use_pi=args.use_pi,
                )
                out1 = hen_root(fname) + "_rms" + ".qdp"
                start_energy = np.asarray(rms.energy_intervals)[:, 0]
                stop_energy = np.asarray(rms.energy_intervals)[:, 1]
                save_as_qdp(
                    [start_energy, stop_energy, rms.spectrum],
                    [None, None, rms.spectrum_error],
                    filename=out1,
                )
                filelist.append(out1)

            if args.lag:
                lag = LagEnergySpectrum(
                    events,
                    args.freq_interval,
                    energy_spec,
                    args.ref_band,
                    segment_size=args.segment_size,
                    bin_time=args.bintime,
                    events2=events2,
                    use_pi=args.use_pi,
                )
                start_energy = np.asarray(lag.energy_intervals)[:, 0]
                stop_energy = np.asarray(lag.energy_intervals)[:, 1]
                out2 = hen_root(fname) + "_lag" + ".qdp"
                save_as_qdp(
                    [start_energy, stop_energy, lag.spectrum],
                    [None, None, lag.spectrum_error],
                    filename=out2,
                )
                filelist.append(out2)

            if args.covariance:
                from stingray.covariancespectrum import (
                    AveragedCovariancespectrum,
                )

                energies = _decode_energy_specification(energy_spec)
                energies = list(zip(energies[:-1], energies[1:]))
                cov = AveragedCovariancespectrum(
                    events,
                    dt=args.bintime,
                    band_interest=energies,
                    ref_band_interest=args.ref_band,
                    segment_size=args.segment_size,
                )

                start_energy = np.asarray(energies)[:, 0]
                stop_energy = np.asarray(energies)[:, 1]
                out2 = hen_root(fname) + "_cov" + ".qdp"
                save_as_qdp(
                    [start_energy, stop_energy, cov.covar],
                    [None, None, cov.covar_error],
                    filename=out2,
                )
                filelist.append(out2)

    return filelist



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test a full run of the codes from the command line."""

import shutil
import os
import glob
import subprocess as sp
import pytest
import numpy as np
from astropy import log
from astropy.io import fits
from astropy.tests.helper import remote_data
import hendrics as hen
from stingray.events import EventList
from hendrics.tests import _dummy_par
from hendrics import fake, calibrate, read_events, io
from hendrics.fake import scramble
from hendrics.io import load_events
from hendrics.fold import HAS_PINT

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

log.setLevel("DEBUG")
# log.basicConfig(filename='HEN.log', level=log.DEBUG, filemode='w')


def test_filter_for_deadtime_nonpar():
    """Test dead time filter, non-paralyzable case."""
    events = np.array([1, 1.05, 1.07, 1.08, 1.1, 2, 2.2, 3, 3.1, 3.2])
    filt_events = hen.fake.filter_for_deadtime(events, 0.11)
    expected = np.array([1, 2, 2.2, 3, 3.2])
    assert np.all(filt_events == expected), "Wrong: {} vs {}".format(
        filt_events, expected
    )


def test_filter_for_deadtime_nonpar_bkg():
    """Test dead time filter, non-paralyzable case, with background."""
    events = np.array([1.1, 2, 2.2, 3, 3.2])
    bkg_events = np.array([1, 3.1])
    filt_events, info = hen.fake.filter_for_deadtime(
        events, 0.11, bkg_ev_list=bkg_events, return_all=True
    )
    expected_ev = np.array([2, 2.2, 3, 3.2])
    expected_bk = np.array([1])
    assert np.all(filt_events == expected_ev), "Wrong: {} vs {}".format(
        filt_events, expected_ev
    )
    assert np.all(info.bkg == expected_bk), "Wrong: {} vs {}".format(
        info.bkg, expected_bk
    )


def test_filter_for_deadtime_par():
    """Test dead time filter, paralyzable case."""
    events = np.array([1, 1.1, 2, 2.2, 3, 3.1, 3.2])
    assert np.all(
        hen.fake.filter_for_deadtime(events, 0.11, paralyzable=True)
        == np.array([1, 2, 2.2, 3])
    )


def test_filter_for_deadtime_par_bkg():
    """Test dead time filter, paralyzable case, with background."""
    events = np.array([1.1, 2, 2.2, 3, 3.2])
    bkg_events = np.array([1, 3.1])
    filt_events, info = hen.fake.filter_for_deadtime(
        events, 0.11, bkg_ev_list=bkg_events, paralyzable=True, return_all=True
    )
    expected_ev = np.array([2, 2.2, 3])
    expected_bk = np.array([1])
    assert np.all(filt_events == expected_ev), "Wrong: {} vs {}".format(
        filt_events, expected_ev
    )
    assert np.all(info.bkg == expected_bk), "Wrong: {} vs {}".format(
        info.bkg, expected_bk
    )


def test_filter_for_deadtime_par_bkg_obj():
    """Test dead time filter on Eventlist, paral. case, with background."""
    times = np.array([1.1, 2, 2.2, 3, 3.2])
    pis = np.arange(times.size) + 5
    energies = np.arange(times.size) + 10
    events = EventList(time=times, energy=energies, pi=pis)
    bkg_events = np.array([1, 3.1])
    filt_events, info = hen.fake.filter_for_deadtime(
        events, 0.11, bkg_ev_list=bkg_events, paralyzable=True, return_all=True
    )
    expected_ev = np.array([2, 2.2, 3])
    expected_bk = np.array([1])
    expected_pi = np.array([6, 7, 8])
    expected_nrg = np.array([11, 12, 13])
    filt_times = filt_events.time
    filt_pis = filt_events.pi
    filt_nrgs = filt_events.energy

    assert np.all(filt_times == expected_ev), "Wrong: {} vs {}".format(
        filt_events, expected_ev
    )
    assert np.all(filt_pis == expected_pi), "Wrong: {} vs {}".format(
        filt_events, expected_ev
    )
    assert np.all(filt_nrgs == expected_nrg), "Wrong: {} vs {}".format(
        filt_events, expected_ev
    )
    assert np.all(info.bkg == expected_bk), "Wrong: {} vs {}".format(
        info.bkg, expected_bk
    )


def test_deadtime_mask_par():
    """Test dead time filter, paralyzable case, with background."""
    events = np.array([1.1, 2, 2.2, 3, 3.2])
    bkg_events = np.array([1, 3.1])
    filt_events, info = hen.fake.filter_for_deadtime(
        events, 0.11, bkg_ev_list=bkg_events, paralyzable=True, return_all=True
    )

    assert np.allclose(filt_events, [2.0, 2.2, 3.0])


def test_deadtime_conversion():
    """Test the functions for count rate conversion."""
    original_rate = np.arange(1, 1000, 10)
    deadtime = 2.5e-3
    rdet = hen.base.r_det(deadtime, original_rate)
    rin = hen.base.r_in(deadtime, rdet)
    np.testing.assert_almost_equal(rin, original_rate)


def verify_all_checksums(filename):
    with fits.open(filename) as hdul:
        for hdu in hdul:
            assert hdu.verify_datasum() == 1, f"Bad datasum: {hdu.name}"
            assert hdu.verify_checksum() == 1, f"Bad checksum: {hdu.name}"


class TestFake(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.first_event_file = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.par = _dummy_par("bubububu.par")
        cls.fits_fileA = os.path.join(cls.datadir, "monol_testA.evt")
        command = "{0} --discard-calibration".format(cls.fits_fileA)
        hen.read_events.main(command.split())

        cls.first_event_file_cal = "calibrated" + HEN_FILE_EXTENSION
        hen.calibrate.calibrate(
            cls.first_event_file, cls.first_event_file_cal, rough=True
        )

        cls.xmm_fits_file = os.path.join(
            cls.datadir, "monol_test_fake_lc_xmm.evt"
        )
        # Note that I don't specify the instrument. This is because
        # I want the internal machinery to understand that this is
        # XMM and this has to be given EPIC-pn by default.
        hen.fake.main(
            [
                "--deadtime",
                "1e-4",
                "-m",
                "XMM",
                "--ctrate",
                "2000",
                "--mjdref",
                "50814.0",
                "-o",
                cls.xmm_fits_file,
            ]
        )
        command = "{0}  --discard-calibration".format(cls.xmm_fits_file)
        hen.read_events.main(command.split())
        cls.xmm_ev_file = os.path.join(
            cls.datadir,
            "monol_test_fake_xmm_xmm_epn_det01_ev" + HEN_FILE_EXTENSION,
        )

    def test_checksums(self):
        for fname in [self.xmm_fits_file, self.fits_fileA]:
            verify_all_checksums(fname)

    def test_fake_file(self):
        """Test produce a fake event file."""
        fits_file = os.path.join(self.datadir, "monol_test_fake.evt")
        hen.fake.main(["-o", fits_file, "--instrument", "FPMB"])
        verify_all_checksums(fits_file)
        info = hen.io.print_fits_info(fits_file, hdu=1)
        assert info["Instrument"] == "FPMB"

    def test_fake_file_from_input_lc(self):
        """Test produce a fake event file from input light curve."""
        lcurve_in = os.path.join(self.datadir, "lcurveA.fits")
        fits_file = os.path.join(self.datadir, "monol_test_fake_lc.evt")
        with pytest.warns(UserWarning) as record:
            hen.fake.main(["--lc", lcurve_in, "-o", fits_file])
        assert np.any(
            [
                "FITS light curve handling is st" in r.message.args[0]
                for r in record
            ]
        )
        verify_all_checksums(fits_file)

    def test_fake_file_with_deadtime(self):
        """Test produce a fake event file and apply deadtime."""
        fits_file = os.path.join(self.datadir, "monol_test_fake_lc.evt")
        hen.fake.main(
            ["--deadtime", "2.5e-3", "--ctrate", "2000", "-o", fits_file]
        )
        verify_all_checksums(fits_file)

    def test_fake_file_xmm(self):
        """Test produce a fake event file and apply deadtime."""

        with fits.open(self.xmm_fits_file) as hdu_list:
            hdunames = [hdu.name for hdu in hdu_list]
            assert "STDGTI01" in hdunames
            assert "STDGTI02" in hdunames
            assert "STDGTI07" in hdunames
            assert hdu_list[0].header["TELESCOP"].lower() == "xmm"
            assert hdu_list[0].header["INSTRUME"].lower() == "epn"

        assert os.path.exists(self.xmm_ev_file)

    def test_load_events_randomize(self):
        """Test event file reading."""
        newfiles = hen.read_events.treat_event_file(
            self.fits_fileA, randomize_by=0.073
        )
        clean_file = self.first_event_file
        ev_clean = hen.io.load_events(clean_file)
        ev = hen.io.load_events(newfiles[0])
        diff = ev.time - ev_clean.time
        assert np.all(np.abs(diff) <= 0.073 / 2)
        assert np.all(np.abs(diff) > 0.0)

    def test_scramble_events_file(self):
        command = f"{self.first_event_file}"
        newfile = hen.fake.main_scramble(command.split())
        assert os.path.exists(newfile)
        os.remove(newfile)

    @pytest.mark.parametrize("fname", ["first_event_file", "xmm_ev_file"])
    def test_fake_fits_input_events_file(self, fname):
        newfile = "bububuasdf.fits"
        infname = getattr(self, fname)
        command = f"-e {infname} -o {newfile}"
        _ = hen.fake.main(command.split())
        assert os.path.exists(newfile)

        verify_all_checksums(newfile)

        events0 = load_events(infname)
        newf = hen.read_events.treat_event_file(newfile)

        events1 = load_events(newf[0])

        assert np.allclose(events0.gti, events1.gti)
        assert np.allclose(events0.time, np.sort(events0.time))
        assert np.allclose(events1.time, np.sort(events1.time))
        assert np.allclose(events0.time, events1.time)

        assert np.isclose(events0.mjdref, events1.mjdref)
        if hasattr(events0, "detector_id") and events0.detector_id is not None:
            assert np.allclose(events0.detector_id, events1.detector_id)

        if "xmm" in fname:
            assert np.isclose(events0.mjdref, 50814)
        os.remove(newfile)

    def test_scramble_uncalibrated_events_file_raises(self):
        command = f"{self.first_event_file} -e 3 30"
        with pytest.raises(ValueError):
            with pytest.warns(UserWarning) as record:
                _ = hen.fake.main_scramble(command.split())
        assert np.any(
            ["No energy information" in r.message.args[0] for r in record]
        )

    def test_scramble_calibrated_events_file(self):
        command = f"{self.first_event_file_cal} -e 3 30"
        newfile = hen.fake.main_scramble(command.split())
        assert "3-30" in newfile
        assert os.path.exists(newfile)
        os.remove(newfile)

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_scramble_events_file_deorbit(self):
        _ = _dummy_par("bububububu.par", pb=1.0, a1=30)
        command = f"{self.first_event_file} --deorbit-par bububububu.par"
        newfile = hen.fake.main_scramble(command.split())
        assert os.path.exists(newfile)
        os.remove(newfile)

    def test_scramble_events(self):
        nevents = 3003
        times = np.random.uniform(0, 1000, nevents)
        times = times[(times > 125.123)]
        # Put exactly one photon inside a very short GTI
        times[0] = 0.5
        times = np.sort(times)
        event_list = EventList(
            times, gti=np.array([[0, 0.9], [111, 123.2], [125.123, 1000]])
        )

        new_event_list = scramble(event_list, "smooth")
        assert new_event_list.time.size == times.size
        assert np.all(new_event_list.gti == event_list.gti)

        new_event_list = scramble(event_list, "flat")
        assert new_event_list.time.size == times.size
        np.all(new_event_list.gti == event_list.gti)

    def test_calibrate_xmm(self):
        """Test event file calibration."""
        xmm_file = self.xmm_ev_file
        command = "{0} -r {1} --nproc 2".format(
            xmm_file, os.path.join(self.datadir, "test.rmf")
        )
        hen.calibrate.main(command.split())

    def test_calibrate_xmm_normf(self):
        """Test event file calibration."""
        xmm_file = self.xmm_ev_file
        command = "{0} --rough --nproc 2".format(xmm_file)
        hen.calibrate.main(command.split())

    @classmethod
    def teardown_class(self):
        """Test a full run of the scripts (command lines)."""

        def find_file_pattern_in_dir(pattern, directory):
            return glob.glob(os.path.join(directory, pattern))

        patterns = [
            "*monol_test*" + HEN_FILE_EXTENSION,
            "*lcurve*" + HEN_FILE_EXTENSION,
            "*lcurve*.txt",
            "*.log",
            "*monol_test*.dat",
            "*monol_test*.png",
            "*monol_test*.txt",
            "*monol_test_fake*.evt",
            "*bubu*",
            "*.p",
            "*.qdp",
            "*.inf",
        ]

        file_list = []
        for pattern in patterns:
            file_list.extend(find_file_pattern_in_dir(pattern, self.datadir))

        for f in file_list:
            if os.path.exists(f):
                print("Removing " + f)
                os.remove(f)

        patterns = ["*_pds*/", "*_cpds*/", "*_sum/"]

        dir_list = []
        for pattern in patterns:
            dir_list.extend(find_file_pattern_in_dir(pattern, self.datadir))
        for f in dir_list:
            if os.path.exists(f):
                shutil.rmtree(f)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to save data in a Xspec-readable format."""

import subprocess as sp
import numpy as np
from astropy import log
from .io import get_file_type
from .io import get_file_extension


def save_as_xspec(fname, direct_save=False, save_lags=True):
    """Save frequency spectra in a format readable to FTOOLS and Xspec.

    Parameters
    ----------
    fname : str
        Input HENDRICS frequency spectrum file name
    direct_save : bool
        If True, also call `flx2xsp` to produce the output .pha and .rsp files.
        If False (default), flx2xsp has to be called from the user

    Notes
    -----
    Uses method described here:
    https://asd.gsfc.nasa.gov/XSPECwiki/fitting_timing_power_spectra_in_XSPEC
    """
    ftype, contents = get_file_type(fname)

    outroot = fname.replace(get_file_extension(fname), "")
    outname = outroot + "_xsp.dat"
    outroot_lags = outroot + "_lags"
    outname_lags = outroot_lags + "_xsp.dat"

    if ftype.endswith("pds"):
        flo = contents.freq - contents.df / 2
        fhi = contents.freq + contents.df / 2
        power = contents.power.real * contents.df
        power_err = contents.power_err.real * contents.df
    else:
        raise ValueError("Data type not supported for Xspec")

    np.savetxt(outname, np.transpose([flo, fhi, power, power_err]))
    if direct_save:
        sp.check_call(
            "flx2xsp {0} {1}.pha {1}.rsp".format(outname, outroot).split()
        )

    if save_lags and ftype == "cpds":
        lags, lags_err = contents.time_lag()
        np.savetxt(
            outname_lags,
            np.transpose(
                [flo, fhi, lags * contents.df, lags_err * contents.df]
            ),
        )
        if direct_save:
            sp.check_call(
                "flx2xsp {0} {1}.pha {1}.rsp".format(
                    outname_lags, outroot_lags
                ).split()
            )


def main(args=None):
    """Main function called by the `HEN2xspec` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Save a frequency spectrum in a qdp file that can be "
        "read by flx2xsp and produce a XSpec-compatible spectrum"
        "file"
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")

    parser.add_argument(
        "--flx2xsp",
        help="Also call flx2xsp at the end",
        default=False,
        action="store_true",
    )
    _add_default_args(parser, ["loglevel", "debug"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)
    files = args.files
    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HEN2Xspec.log"):
        for f in files:
            save_as_xspec(f, direct_save=args.flx2xsp)



import os
from astropy.io import fits
import numpy as np
import pytest
from ..phasetag import main_phasetag


class TestPhasetag:
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.fits_fileA = os.path.join(cls.datadir, "monol_testA.evt")
        cls.freq = 0.1235242

    @pytest.mark.parametrize("N", [5, 6, 11, 16, 32, 41])
    def test_phase_tag(self, N):
        main_phasetag(
            [
                self.fits_fileA,
                "-f",
                str(self.freq),
                "--test",
                "--tomax",
                "-n",
                str(N),
                "--plot",
            ]
        )
        self.phasetagged = self.fits_fileA.replace(".evt", "_phasetag.evt")
        assert os.path.exists(self.phasetagged)

        # Redo to test if existing columns are preserved
        main_phasetag(
            [
                self.phasetagged,
                "-f",
                str(self.freq),
                "--test",
                "--tomax",
                "-n",
                str(N),
            ]
        )

        hdulist = fits.open(self.phasetagged)
        for hdu in hdulist:
            assert hdu.verify_checksum() == 1
            assert hdu.verify_datasum() == 1

        times = np.array(hdulist[1].data["TIME"])
        phases = np.array(hdulist[1].data["Phase"])
        orbtimes = np.array(hdulist[1].data["Orbit_bary"])
        hdulist.close()

        assert np.all(phases >= 0)
        assert np.all(phases < 1)
        # There's no orbit in these files
        assert np.allclose(times, orbtimes, rtol=1e-15)

        phase_bins = np.linspace(0, 1, N + 1)
        prof, _ = np.histogram(phases, bins=phase_bins)

        # I used --tomax
        assert np.argmax(prof) == 0
        os.unlink(self.phasetagged)

    @pytest.mark.parametrize("N", [5, 6, 11, 16, 32, 41])
    def test_phase_tag_TOA(self, N):
        main_phasetag(
            [
                self.fits_fileA,
                "-f",
                str(self.freq),
                "--test",
                "--refTOA",
                "57000.01",
                "-n",
                str(N),
                "--plot",
            ]
        )
        self.phasetagged = self.fits_fileA.replace(".evt", "_phasetag.evt")
        assert os.path.exists(self.phasetagged)

    def test_phase_tag_badexposure(self):
        with pytest.warns(UserWarning) as record:
            main_phasetag(
                [
                    self.fits_fileA,
                    "-f",
                    "0.00001",
                    "--test",
                    "--tomax",
                    "-n",
                    "1000",
                    "--plot",
                ]
            )
        assert np.any(
            [
                "Exposure has NaNs or zeros. " in r.message.args[0]
                for r in record
            ]
        )

    def test_phase_tag_invalid0(self):
        with pytest.raises(ValueError) as excinfo:
            main_phasetag([self.fits_fileA, "--test"])
        assert "Specify one between" in str(excinfo.value)

    def test_phase_tag_invalid1(self):
        with pytest.raises(ValueError) as excinfo:
            main_phasetag(
                [self.fits_fileA, "-f", "1", "--parfile", "bubu.par", "--test"]
            )
        assert "Specify only one between" in str(excinfo.value)

    def test_phase_tag_parfile(self):
        with pytest.raises(NotImplementedError) as excinfo:
            main_phasetag([self.fits_fileA, "--parfile", "bubu.par", "--test"])
        assert "This part is not yet implemented" in str(excinfo.value)
